#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'fci_l1c_nc' reader."""
import contextlib
import datetime
import logging
import os
from typing import Dict, List, Union
from unittest import mock

import dask.array as da
import numpy as np
import numpy.testing
import pytest
import xarray as xr
from netCDF4 import default_fillvals
from pytest_lazy_fixtures import lf as lazy_fixture

from satpy.readers.fci_l1c_nc import FCIL1cNCFileHandler
from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler
from satpy.tests.utils import make_dataid

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - caplog

GRID_TYPE_INFO_FOR_TEST_CONTENT = {
    "500m": {
        "nrows": 400,
        "ncols": 22272,
        "scale_factor": 1.39717881644274e-05,
        "add_offset": 1.55596818893146e-01,
    },
    "1km": {
        "nrows": 200,
        "ncols": 11136,
        "scale_factor": 2.79435763233999e-05,
        "add_offset": 1.55603804756852e-01,
    },
    "2km": {
        "nrows": 100,
        "ncols": 5568,
        "scale_factor": 5.58871526031607e-05,
        "add_offset": 1.55617776423501e-01,
    },
    "3km": {
        "nrows": 67,
        "ncols": 3712,
        "scale_factor": 8.38307287956433e-05,
        "add_offset": 0.155631748009112,
    },
}

LIST_CHANNEL_SOLAR = ["vis_04", "vis_05", "vis_06", "vis_08", "vis_09",
                      "nir_13", "nir_16", "nir_22"]
LIST_CHANNEL_TERRAN = ["ir_38", "wv_63", "wv_73", "ir_87", "ir_97", "ir_105",
                       "ir_123", "ir_133"]
LIST_TOTAL_CHANNEL = LIST_CHANNEL_SOLAR + LIST_CHANNEL_TERRAN
LIST_RESOLUTION_VIS06_AF = ["1km", "3km"]
LIST_RESOLUTION_AF = ["3km"]
EXPECTED_POS_INFO_FOR_FILETYPE = {
    "fdhsi": {"1km": {"start_position_row": 1,
                      "end_position_row": 200,
                      "segment_height": 200,
                      "grid_width": 11136},
              "2km": {"start_position_row": 1,
                      "end_position_row": 100,
                      "segment_height": 100,
                      "grid_width": 5568}},
    "hrfi": {"500m": {"start_position_row": 1,
                      "end_position_row": 400,
                      "segment_height": 400,
                      "grid_width": 22272},
             "1km": {"start_position_row": 1,
                     "end_position_row": 200,
                     "grid_width": 11136,
                     "segment_height": 200}},
    "fci_af": {"3km": {"start_position_row": 1,
                       "end_position_row": 67,
                       "segment_height": 67,
                       "grid_width": 3712
                       },
               },
    "fci_af_vis_06": {"3km": {"start_position_row": 1,
                              "end_position_row": 67,
                              "segment_height": 67,
                              "grid_width": 3712
                              },
                      "1km": {"start_position_row": 1,
                              "end_position_row": 200,
                              "grid_width": 11136,
                              "segment_height": 200}
                      }
}

CHANS_FDHSI = {"solar": LIST_CHANNEL_SOLAR,
               "solar_grid_type": ["1km"] * 8,
               "terran": LIST_CHANNEL_TERRAN,
               "terran_grid_type": ["2km"] * 8}

CHANS_HRFI = {"solar": ["vis_06", "nir_22"],
              "solar_grid_type": ["500m"] * 2,
              "terran": ["ir_38", "ir_105"],
              "terran_grid_type": ["1km"] * 2}

DICT_CALIBRATION = {"radiance": {"dtype": np.float32,
                                 "value_1": 15,
                                 "value_0": 9700,
                                 "value_2": -5,
                                 "attrs_dict": {"calibration": "radiance",
                                                "units": "mW m-2 sr-1 (cm-1)-1",
                                                "radiance_unit_conversion_coefficient": np.float32(1234.56)
                                                },
                                 },

                    "reflectance": {"dtype": np.float32,
                                    "attrs_dict": {"calibration": "reflectance",
                                                   "units": "%"
                                                   },
                                    },

                    "counts": {"dtype": np.uint16,
                               "value_1": 5,
                               "value_0": 5000,
                               "value_2": 1,
                               "attrs_dict": {"calibration": "counts",
                                              "units": "count",
                                              },
                               },

                    "brightness_temperature": {"dtype": np.float32,
                                               "value_1": np.float32(209.68275),
                                               "value_0": np.float32(1888.8513),
                                               "value_2": np.float32("nan"),
                                               "attrs_dict": {"calibration": "brightness_temperature",
                                                              "units": "K",
                                                              },
                                               },
                    }
TEST_FILENAMES = {"fdhsi": [
    "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
    "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
    "20170410113925_20170410113934_N__C_0070_0067.nc"
],
    "fdhsi_iqti": [
        "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--"
        "CHK-BODY--MON-NC4_C_EUMT_20240307233956_IQTI_DEV_"
        "20231016125007_20231016125017_N__C_0078_0001.nc"
    ],
    "hrfi": [
        "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-HRFI-FD--"
        "CHK-BODY--L2P-NC4E_C_EUMT_20170410114434_GTT_DEV_"
        "20170410113925_20170410113934_N__C_0070_0067.nc"
    ],
    "hrfi_iqti": [
        "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-HRFI-FD--"
        "CHK-BODY--MON-NC4_C_EUMT_20240307233956_IQTI_DEV_"
        "20231016125007_20231016125017_N__C_0078_0001.nc"
    ],
    "fdhsi_q4": ["W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-Q4--"
                 "CHK-BODY--DIS-NC4E_C_EUMT_20230723025408_IDPFI_DEV_"
                 "20230722120000_20230722120027_N_JLS_C_0289_0001.nc"
                 ],
    "hrfi_q4": ["W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-HRFI-Q4--"
                "CHK-BODY--DIS-NC4E_C_EUMT_20230723025408_IDPFI_DEV"
                "_20230722120000_20230722120027_N_JLS_C_0289_0001.nc"],
    "hrfi_other_coverage": ["W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-HRFI-xx--"
                "CHK-BODY--DIS-NC4E_C_EUMT_20230723025408_IDPFI_DEV"
                "_20230722120000_20230722120027_N_JLS_C_0289_0001.nc"],
    "fdhsi_other_coverage": ["W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-xx--"
                 "CHK-BODY--DIS-NC4E_C_EUMT_20230723025408_IDPFI_DEV_"
                 "20230722120000_20230722120027_N_JLS_C_0289_0001.nc"
                 ],
}

REQUIRED_ATTR_KEYS = {
    "time_parameters",
    "orbital_parameters",
    "file_type",
    "name",
    "sensor",
    "resolution",
}

def resolutions_AF_products(channel):
    """Get the resolutions of the African products."""
    if channel == "vis_06":
        return LIST_RESOLUTION_VIS06_AF
    else:
        return LIST_RESOLUTION_AF


def fill_chans_af():
    """Fill the dict CHANS_AF and the list TEST_FILENAMES with the right channel and resolution."""
    chans_af = {}
    for channel in LIST_TOTAL_CHANNEL:
        list_resol = resolutions_AF_products(channel)
        if channel in ["wv_63", "wv_73"]:
            ch_name_for_file = channel.replace("wv", "ir").replace("_", "").upper()
        else:
            ch_name_for_file = channel.replace("_", "").upper()
        for resol in list_resol:
            TEST_FILENAMES[f"af_{channel}_{resol}"] = [f"W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1-FCI-1C-RRAD"
                                                       f"-{resol.upper()}-AF-{ch_name_for_file}-x-x---NC4E_C_EUMT_"
                                                       f"20240125144655_DT_OPE"
                                                       f"_20240109080007_20240109080924_N_JLS_T_0049_0000.nc"]
            if channel.split("_")[0] in ["vis", "nir"]:
                chans_af[f"{channel}_{resol}"] = {"solar": [channel],
                                                  "solar_grid_type": [resol]}
            elif channel.split("_")[0] in ["ir", "wv"]:
                chans_af[f"{channel}_{resol}"] = {"terran": [channel],
                                                  "terran_grid_type": [resol]}
    return chans_af


CHANS_AF = fill_chans_af()


# ----------------------------------------------------
# Filehandlers preparation ---------------------------
# ----------------------------------------------------

class FakeH5Variable:
    """Class for faking h5netcdf.Variable class."""

    def __init__(self, data, dims=(), attrs=None):
        """Initialize the class."""
        self.dimensions = dims
        self.name = "name"
        self.attrs = attrs if attrs else {}
        self.dtype = None
        self._data = data
        self._set_meta()

    def _set_meta(self):
        if hasattr(self._data, "dtype"):
            self.dtype = self._data.dtype

    def __array__(self):
        """Get the array data."""
        return self._data.__array__()

    def __getitem__(self, key):
        """Get item for the key."""
        return self._data[key]

    @property
    def shape(self):
        """Get the shape."""
        return self._data.shape

    @property
    def ndim(self):
        """Get the number of dimensions."""
        return self._data.ndim


def _get_test_calib_for_channel_ir(data, meas_path):
    from pyspectral.blackbody import C_SPEED as c
    from pyspectral.blackbody import H_PLANCK as h
    from pyspectral.blackbody import K_BOLTZMANN as k
    data[meas_path + "/radiance_to_bt_conversion_coefficient_wavenumber"] = FakeH5Variable(
        da.array(955.0, dtype=np.float32))
    data[meas_path + "/radiance_to_bt_conversion_coefficient_a"] = FakeH5Variable(da.array(1.0, dtype=np.float32))
    data[meas_path + "/radiance_to_bt_conversion_coefficient_b"] = FakeH5Variable(da.array(0.4, dtype=np.float32))
    data[meas_path + "/radiance_to_bt_conversion_constant_c1"] = FakeH5Variable(
        da.array(1e11 * 2 * h * c ** 2, dtype=np.float32))
    data[meas_path + "/radiance_to_bt_conversion_constant_c2"] = FakeH5Variable(
        da.array(1e2 * h * c / k, dtype=np.float32))
    return data


def _get_test_calib_for_channel_vis(data, meas):
    data["state/celestial/earth_sun_distance"] = FakeH5Variable(
        da.repeat(da.array([149597870.7]), 6000), dims="index")
    data[meas + "/channel_effective_solar_irradiance"] = FakeH5Variable(da.array(50.0, dtype=np.float32))
    return data


def _get_test_calib_data_for_channel(data, ch_str):
    meas_path = "data/{:s}/measured".format(ch_str)
    if ch_str.startswith("ir") or ch_str.startswith("wv"):
        _get_test_calib_for_channel_ir(data, meas_path)
    elif ch_str.startswith("vis") or ch_str.startswith("nir"):
        _get_test_calib_for_channel_vis(data, meas_path)
    data[meas_path + "/radiance_unit_conversion_coefficient"] = xr.DataArray(np.array(1234.56, dtype=np.float32))


def _get_test_image_data_for_channel(data, ch_str, n_rows_cols):
    ch_path = "data/{:s}/measured/effective_radiance".format(ch_str)

    common_attrs = {
        "scale_factor": 5,
        "add_offset": -10,
        "long_name": "Effective Radiance",
        "units": "mW.m-2.sr-1.(cm-1)-1",
        "ancillary_variables": "pixel_quality",
        "_FillValue": 65535
    }
    if "38" in ch_path:
        fire_line = da.ones((1, n_rows_cols[1]), dtype="uint16", chunks=1024) * 5000
        data_without_fires = da.full((n_rows_cols[0] - 2, n_rows_cols[1]), 5, dtype="uint16", chunks=1024)
        neg_rad = da.ones((1, n_rows_cols[1]), dtype="uint16", chunks=1024)
        d = FakeH5Variable(
            da.concatenate([fire_line, data_without_fires, neg_rad], axis=0),
            dims=("y", "x"),
            attrs={
                "valid_range": [0, 8191],
                "warm_scale_factor": np.float32(2.0),
                "warm_add_offset": np.float32(-300.0),
                **common_attrs
            }
        )
    else:
        d = FakeH5Variable(
            da.full(n_rows_cols, 5, dtype="uint16", chunks=1024),
            dims=("y", "x"),
            attrs={
                "valid_range": [0, 4095],
                "warm_scale_factor": np.float32(1.0),
                "warm_add_offset": np.float32(0.0),
                **common_attrs
            }
        )

    data[ch_path] = d
    data[ch_path + "/shape"] = n_rows_cols


def _get_test_segment_position_for_channel(data, ch_str, n_rows_cols):
    pos = "data/{:s}/measured/{:s}_position_{:s}"
    data[pos.format(ch_str, "start", "row")] = FakeH5Variable(da.array(1))
    data[pos.format(ch_str, "start", "column")] = FakeH5Variable(da.array(1))
    data[pos.format(ch_str, "end", "row")] = FakeH5Variable(da.array(n_rows_cols[0]))
    data[pos.format(ch_str, "end", "column")] = FakeH5Variable(da.array(n_rows_cols[1]))


def _get_test_index_map_for_channel(data, ch_str, n_rows_cols):
    index_map_path = "data/{:s}/measured/index_map".format(ch_str)
    data[index_map_path] = xr.DataArray(
        (da.ones(n_rows_cols)) * 110, dims=("y", "x"))


def _get_test_pixel_quality_for_channel(data, ch_str, n_rows_cols):
    qual_path = "data/{:s}/measured/pixel_quality".format(ch_str)
    data[qual_path] = xr.DataArray((da.ones(n_rows_cols)) * 3, dims=("y", "x"))


def _get_test_geolocation_for_channel(data, ch_str, grid_type, n_rows_cols):
    x_path = "data/{:s}/measured/x".format(ch_str)
    data[x_path] = xr.DataArray(
        da.arange(1, n_rows_cols[1] + 1, dtype=np.dtype("uint16")),
        dims=("x",),
        attrs={
            "scale_factor": -GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["scale_factor"],
            "add_offset": GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["add_offset"],
        }
    )

    y_path = "data/{:s}/measured/y".format(ch_str)
    data[y_path] = xr.DataArray(
        da.arange(1, n_rows_cols[0] + 1, dtype=np.dtype("uint16")),
        dims=("y",),
        attrs={
            "scale_factor": GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["scale_factor"],
            "add_offset": -GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["add_offset"],
        }
    )


def _get_test_content_areadef():
    data = {}

    proj = "data/mtg_geos_projection"

    attrs = {
        "sweep_angle_axis": "y",
        "perspective_point_height": "35786400.0",
        "semi_major_axis": "6378137.0",
        "longitude_of_projection_origin": "0.0",
        "inverse_flattening": "298.257223563",
        "units": "m"}
    data[proj] = xr.DataArray(
        0,
        dims=(),
        attrs=attrs)

    # also set attributes cached, as this may be how they are accessed with
    # the NetCDF4FileHandler
    for (k, v) in attrs.items():
        data[proj + "/attr/" + k] = v

    return data


def _get_test_content_aux_data():
    from satpy.readers.fci_l1c_nc import AUX_DATA
    data = {}
    indices_dim = 6000
    for key, value in AUX_DATA.items():
        # skip population of earth_sun_distance as this is already defined for reflectance calculation
        if key == "earth_sun_distance":
            continue
        data[value] = xr.DataArray(da.arange(indices_dim, dtype=np.dtype("float32")), dims="index")

    # compute the last data entry to simulate the FCI caching
    # data[list(AUX_DATA.values())[-1]] = data[list(AUX_DATA.values())[-1]].compute()

    data["index"] = xr.DataArray(
        da.ones(indices_dim, dtype="uint16") * 100, dims="index")
    return data


def _get_global_attributes():
    data = {}
    attrs = {"platform": "MTI1"}
    for (k, v) in attrs.items():
        data["attr/" + k] = v
    return data


def _get_test_content_for_channel(ch_str, grid_type):
    nrows = GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["nrows"]
    ncols = GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["ncols"]
    n_rows_cols = (nrows, ncols)

    data = {}

    _get_test_image_data_for_channel(data, ch_str, n_rows_cols)
    _get_test_calib_data_for_channel(data, ch_str)
    _get_test_geolocation_for_channel(data, ch_str, grid_type, n_rows_cols)
    _get_test_pixel_quality_for_channel(data, ch_str, n_rows_cols)
    _get_test_index_map_for_channel(data, ch_str, n_rows_cols)
    _get_test_segment_position_for_channel(data, ch_str, n_rows_cols)

    return data


class FakeFCIFileHandlerBase(FakeNetCDF4FileHandler):
    """Class for faking the NetCDF4 Filehandler."""

    cached_file_content: Dict[str, xr.DataArray] = {}
    # overwritten by FDHSI and HRFI File Handlers
    chan_patterns: Dict[str, Dict[str, Union[List[int], str]]] = {}

    def __init__(self, *args, **kwargs):
        """Initiative fake file handler."""
        kwargs.pop("clip_negative_radiances", None)
        super().__init__(*args, **kwargs)

    def _get_test_content_all_channels(self):
        data = {}
        for pat in self.chan_patterns:
            for ch in self.chan_patterns[pat]["channels"]:
                data.update(_get_test_content_for_channel(pat.format(ch), self.chan_patterns[pat]["grid_type"]))
        return data

    def get_test_content(self, filename, filename_info, filetype_info):
        """Get the content of the test data."""
        D = {}
        D.update(self._get_test_content_all_channels())
        D.update(_get_test_content_areadef())
        D.update(_get_test_content_aux_data())
        D.update(_get_global_attributes())
        return D


class FakeFCIFileHandlerFDHSI(FakeFCIFileHandlerBase):
    """Mock FDHSI data."""

    chan_patterns = {
        "vis_{:>02d}": {"channels": [4, 5, 6, 8, 9],
                        "grid_type": "1km"},
        "nir_{:>02d}": {"channels": [13, 16, 22],
                        "grid_type": "1km"},
        "ir_{:>02d}": {"channels": [38, 87, 97, 105, 123, 133],
                       "grid_type": "2km"},
        "wv_{:>02d}": {"channels": [63, 73],
                       "grid_type": "2km"},
    }


class FakeFCIFileHandlerFDHSIIQTI(FakeFCIFileHandlerFDHSI):
    """Mock IQTI for FHDSI data."""

    def _get_test_content_all_channels(self):
        data = super()._get_test_content_all_channels()
        data.update({"state/celestial/earth_sun_distance": FakeH5Variable(
            da.repeat(da.array([np.nan]), 6000), dims="index")})
        return data


class FakeFCIFileHandlerWithBadData(FakeFCIFileHandlerFDHSI):
    """Mock bad data."""

    def _get_test_content_all_channels(self):
        data = super()._get_test_content_all_channels()
        v = xr.DataArray(default_fillvals["f4"])

        data.update({"data/ir_105/measured/radiance_to_bt_conversion_coefficient_wavenumber": v,
                     "data/ir_105/measured/radiance_to_bt_conversion_coefficient_a": v,
                     "data/ir_105/measured/radiance_to_bt_conversion_coefficient_b": v,
                     "data/ir_105/measured/radiance_to_bt_conversion_constant_c1": v,
                     "data/ir_105/measured/radiance_to_bt_conversion_constant_c2": v,
                     "data/vis_06/measured/channel_effective_solar_irradiance": v})

        return data


class FakeFCIFileHandlerHRFI(FakeFCIFileHandlerBase):
    """Mock HRFI data."""

    chan_patterns = {
        "vis_{:>02d}_hr": {"channels": [6],
                           "grid_type": "500m"},
        "nir_{:>02d}_hr": {"channels": [22],
                           "grid_type": "500m"},
        "ir_{:>02d}_hr": {"channels": [38, 105],
                          "grid_type": "1km"},
    }


class FakeFCIFileHandlerHRFIIQTI(FakeFCIFileHandlerHRFI):
    """Mock IQTI for HRFI data."""

    def _get_test_content_all_channels(self):
        data = super()._get_test_content_all_channels()
        data.update({"state/celestial/earth_sun_distance": FakeH5Variable(
            da.repeat(da.array([np.nan]), 6000), dims="x")})
        return data


class FakeFCIFileHandlerAF(FakeFCIFileHandlerBase):
    """Mock AF data."""
    chan_patterns = {}


# ----------------------------------------------------
# Fixtures preparation -------------------------------
# ----------------------------------------------------

@pytest.fixture
def reader_configs():
    """Return reader configs for FCI."""
    from satpy._config import config_search_paths
    return config_search_paths(
        os.path.join("readers", "fci_l1c_nc.yaml"))


def _get_reader_with_filehandlers(filenames, reader_configs, **reader_kwargs):
    from satpy.readers.core.loading import load_reader
    reader = load_reader(reader_configs)
    loadables = reader.select_files_from_pathnames(filenames)
    reader.create_filehandlers(loadables, fh_kwargs=reader_kwargs)
    clear_cache(reader)
    return reader


def clear_cache(reader):
    """Clear the cache for file handlres in reader."""
    for key in reader.file_handlers:
        fhs = reader.file_handlers[key]
        for fh in fhs:
            fh.cached_file_content = {}


def get_list_channel_calibration(calibration):
    """Get the channel's list according the calibration."""
    if calibration == "reflectance":
        return LIST_CHANNEL_SOLAR
    elif calibration == "brightness_temperature":
        return LIST_CHANNEL_TERRAN
    else:
        return LIST_TOTAL_CHANNEL


def generate_parameters(calibration):
    """Generate dynamically the parameters."""
    for channel in get_list_channel_calibration(calibration):
        for resolution in resolutions_AF_products(channel):
            yield channel, resolution


@contextlib.contextmanager
def mocked_basefilehandler(filehandler):
    """Mock patch the base class of the FCIL1cNCFileHandler with the content of our fake files (filehandler)."""
    p = mock.patch.object(FCIL1cNCFileHandler, "__bases__", (filehandler,))
    with p:
        p.is_local = True
        yield


@pytest.fixture
def FakeFCIFileHandlerFDHSI_fixture():
    """Get a fixture for the fake FDHSI filehandler, including channel and file names."""
    with mocked_basefilehandler(FakeFCIFileHandlerFDHSI):
        param_dict = {
            "filetype": "fci_l1c_fdhsi",
            "channels": CHANS_FDHSI,
            "filenames": TEST_FILENAMES["fdhsi"]
        }
        yield param_dict


@pytest.fixture
def FakeFCIFileHandlerFDHSIIQTI_fixture():
    """Get a fixture for the fake FDHSI IQTI filehandler, including channel and file names."""
    with mocked_basefilehandler(FakeFCIFileHandlerFDHSIIQTI):
        param_dict = {
            "filetype": "fci_l1c_fdhsi",
            "channels": CHANS_FDHSI,
            "filenames": TEST_FILENAMES["fdhsi_iqti"]
        }
        yield param_dict


@pytest.fixture
def FakeFCIFileHandlerFDHSIOtherCoverage_fixture():
    """Get a fixture for the fake FDHSI filehandler for coverages not defined, including channel and file names."""
    with mocked_basefilehandler(FakeFCIFileHandlerFDHSI):
        param_dict = {
            "filetype": "fci_l1c_fdhsi",
            "channels": CHANS_FDHSI,
            "filenames": TEST_FILENAMES["fdhsi_other_coverage"]
        }
        yield param_dict

@pytest.fixture
def FakeFCIFileHandlerFDHSIQ4_fixture():
    """Get a fixture for the fake FDHSI Q4 filehandler, including channel and file names."""
    with mocked_basefilehandler(FakeFCIFileHandlerFDHSI):
        param_dict = {
            "filetype": "fci_l1c_fdhsi",
            "channels": CHANS_FDHSI,
            "filenames": TEST_FILENAMES["fdhsi_q4"]
        }
        yield param_dict


@pytest.fixture
def FakeFCIFileHandlerHRFI_fixture():
    """Get a fixture for the fake HRFI filehandler, including channel and file names."""
    with mocked_basefilehandler(FakeFCIFileHandlerHRFI):
        param_dict = {
            "filetype": "fci_l1c_hrfi",
            "channels": CHANS_HRFI,
            "filenames": TEST_FILENAMES["hrfi"]
        }
        yield param_dict

@pytest.fixture
def FakeFCIFileHandlerHRFIOtherCoverage_fixture():
    """Get a fixture for the fake HRFI filehandler for coverages not defined, including channel and file names."""
    with mocked_basefilehandler(FakeFCIFileHandlerHRFI):
        param_dict = {
            "filetype": "fci_l1c_hrfi",
            "channels": CHANS_HRFI,
            "filenames": TEST_FILENAMES["hrfi_other_coverage"]
        }
        yield param_dict

@pytest.fixture
def FakeFCIFileHandlerHRFIIQTI_fixture():
    """Get a fixture for the fake HRFI IQTI filehandler, including channel and file names."""
    with mocked_basefilehandler(FakeFCIFileHandlerHRFIIQTI):
        param_dict = {
            "filetype": "fci_l1c_hrfi",
            "channels": CHANS_HRFI,
            "filenames": TEST_FILENAMES["hrfi_iqti"]
        }
        yield param_dict


@pytest.fixture
def FakeFCIFileHandlerHRFIQ4_fixture():
    """Get a fixture for the fake HRFI Q4 filehandler, including channel and file names."""
    with mocked_basefilehandler(FakeFCIFileHandlerHRFI):
        param_dict = {
            "filetype": "fci_l1c_hrfi",
            "channels": CHANS_HRFI,
            "filenames": TEST_FILENAMES["hrfi_q4"]
        }
        yield param_dict


@pytest.fixture
def FakeFCIFileHandlerAF_fixture(channel, resolution):
    """Get a fixture for the fake AF filehandler, it contains only one channel and one resolution."""
    chan_patterns = {channel.split("_")[0] + "_{:>02d}": {"channels": [int(channel.split("_")[1])],
                                                          "grid_type": f"{resolution}"}, }
    FakeFCIFileHandlerAF.chan_patterns = chan_patterns
    with mocked_basefilehandler(FakeFCIFileHandlerAF):
        param_dict = {
            "filetype": "fci_l1c_af",
            "channels": CHANS_AF[f"{channel}_{resolution}"],
            "filenames": TEST_FILENAMES[f"af_{channel}_{resolution}"],
        }
        yield param_dict


# ----------------------------------------------------
# Tests ----------------------------------------------
# ----------------------------------------------------
class ModuleTestFCIL1cNcReader:
    """Class containing parameters and modules useful for the test related to L1c reader."""
    fh_param_for_filetype = {"hrfi": {"channels": CHANS_HRFI,
                                      "filenames": TEST_FILENAMES["hrfi"]},
                             "fdhsi": {"channels": CHANS_FDHSI,
                                       "filenames": TEST_FILENAMES["fdhsi"]},
                             "fdhsi_iqti": {"channels": CHANS_FDHSI,
                                            "filenames": TEST_FILENAMES["fdhsi_iqti"]},
                             "hrfi_q4": {"channels": CHANS_HRFI,
                                         "filenames": TEST_FILENAMES["hrfi_q4"]},
                             "hrfi_iqti": {"channels": CHANS_HRFI,
                                           "filenames": TEST_FILENAMES["hrfi_iqti"]},
                             "fdhsi_q4": {"channels": CHANS_FDHSI,
                                          "filenames": TEST_FILENAMES["fdhsi_q4"]},
                             "fdhsi_other_coverage": {"channels": CHANS_FDHSI,
                                          "filenames": TEST_FILENAMES["fdhsi_other_coverage"]},
                             "hrfi_other_coverage": {"channels": CHANS_HRFI,
                                         "filenames": TEST_FILENAMES["hrfi_other_coverage"]}}

    @staticmethod
    def _get_type_ter_AF(channel):
        """Get the type_ter."""
        if channel.split("_")[0] in ["vis", "nir"]:
            return "solar"
        elif channel.split("_")[0] in ["wv", "ir"]:
            return "terran"

    @staticmethod
    def _get_assert_attrs(res, ch, attrs_dict):
        """Test the different attributes values."""
        for key, item in attrs_dict.items():
            assert res[ch].attrs[key] == item

    @staticmethod
    def _get_assert_erased_attrs(res, ch):
        """Test that the attributes listed have been erased."""
        LIST_ATTRIBUTES = ["add_offset", "warm_add_offset", "scale_factor",
                           "warm_scale_factor", "valid_range"]
        for atr in LIST_ATTRIBUTES:
            assert atr not in res[ch].attrs

    @staticmethod
    def _reflectance_test(tab, filenames):
        """Test of with the reflectance test."""
        if "IQTI" in filenames:
            numpy.testing.assert_array_almost_equal(tab,
                                                    93.6462, decimal=4)
        else:
            numpy.testing.assert_array_almost_equal(tab,
                                                    100 * 15 * 1 * np.pi / 50)

    @staticmethod
    def _other_calibration_test(res, ch, dict_arg):
        """Test of other calibration test."""
        if ch == "ir_38":
            numpy.testing.assert_array_equal(res[ch][-1], dict_arg["value_2"])
            numpy.testing.assert_array_equal(res[ch][-2], dict_arg["value_1"])
            numpy.testing.assert_array_equal(res[ch][0], dict_arg["value_0"])
        else:
            numpy.testing.assert_array_equal(res[ch], dict_arg["value_1"])

    @staticmethod
    def _shape_test(res, ch, grid_type, dict_arg):
        """Test the shape."""
        assert res[ch].shape == (GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["nrows"],
                                 GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["ncols"])
        assert res[ch].dtype == dict_arg["dtype"]

    def _get_assert_load(self, res, ch, dict_arg, filenames):
        """Test the value for different channels."""
        self._get_assert_attrs(res, ch, dict_arg["attrs_dict"])
        if dict_arg["attrs_dict"]["calibration"] in ["radiance", "brightness_temperature", "reflectance"]:
            self._get_assert_erased_attrs(res, ch)
        if dict_arg["attrs_dict"]["calibration"] == "reflectance":
            self._reflectance_test(res[ch], filenames)
        else:
            self._other_calibration_test(res, ch, dict_arg)

    def _get_res_AF(self, channel, fh_param, calibration, reader_configs):
        """Load the reader for AF data."""
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        type_ter = self._get_type_ter_AF(channel)
        res = reader.load([make_dataid(name=name, calibration=calibration)
                           for name in fh_param["channels"][type_ter]], pad_data=False)
        return res

    @staticmethod
    def _compare_sun_earth_distance(filetype, fh_param, reader_configs):
        """Test the sun earth distance calculation."""
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        if "IQTI" in fh_param["filenames"][0]:
            np.testing.assert_almost_equal(
                reader.file_handlers[filetype][0]._compute_sun_earth_distance,
                0.996803423, decimal=7)
        else:
            np.testing.assert_almost_equal(
                reader.file_handlers[filetype][0]._compute_sun_earth_distance,
                1.0, decimal=7)

    @staticmethod
    def _compare_rc_period_min_count_in_repeat_cycle(filetype, fh_param,
                                                     reader_configs, compare_parameters_tuple):
        """Test the count_in_repeat_cycle, rc_period_min."""
        count_in_repeat_cycle_imp, rc_period_min_imp, start_nominal_time, end_nominal_time = compare_parameters_tuple
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        assert count_in_repeat_cycle_imp == \
               reader.file_handlers[filetype][0].filename_info["count_in_repeat_cycle"]
        assert rc_period_min_imp == \
               reader.file_handlers[filetype][0].rc_period_min
        assert start_nominal_time == reader.file_handlers[filetype][0].nominal_start_time
        assert end_nominal_time == reader.file_handlers[filetype][0].nominal_end_time


class TestFCIL1cNCReader(ModuleTestFCIL1cNcReader):
    """Test FCI L1c NetCDF reader with nominal data."""

    @pytest.mark.parametrize("filenames", [TEST_FILENAMES[filename] for filename in TEST_FILENAMES.keys()])
    def test_file_pattern(self, reader_configs, filenames):
        """Test file pattern matching."""
        from satpy.readers.core.loading import load_reader

        reader = load_reader(reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert len(files) == 1

    @pytest.mark.parametrize("filenames", [TEST_FILENAMES["fdhsi"][0].replace("BODY", "TRAIL"),
                                           TEST_FILENAMES["hrfi"][0].replace("BODY", "TRAIL"),
                                           TEST_FILENAMES["hrfi_q4"][0].replace("BODY", "TRAIL"),
                                           TEST_FILENAMES["fdhsi_q4"][0].replace("BODY", "TRAIL"),
                                           TEST_FILENAMES["fdhsi_iqti"][0].replace("BODY", "TRAIL"),
                                           TEST_FILENAMES["hrfi_iqti"][0].replace("BODY", "TRAIL"),
                                           TEST_FILENAMES["hrfi_other_coverage"][0].replace("BODY", "TRAIL"),
                                           TEST_FILENAMES["fdhsi_other_coverage"][0].replace("BODY", "TRAIL")])
    def test_file_pattern_for_TRAIL_file(self, reader_configs, filenames):
        """Test file pattern matching for TRAIL files, which should not be picked up."""
        from satpy.readers.core.loading import load_reader

        reader = load_reader(reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert len(files) == 0

    @pytest.mark.parametrize("calibration", ["counts", "radiance", "brightness_temperature", "reflectance"])
    @pytest.mark.parametrize(("fh_param", "res_type"), [
            (lazy_fixture("FakeFCIFileHandlerFDHSI_fixture"), "hdfi"),
            (lazy_fixture("FakeFCIFileHandlerHRFI_fixture"), "hrfi"),
            (lazy_fixture("FakeFCIFileHandlerHRFIQ4_fixture"), "hrfi"),
            (lazy_fixture("FakeFCIFileHandlerFDHSIQ4_fixture"), "hdfi"),
            (lazy_fixture("FakeFCIFileHandlerHRFIIQTI_fixture"), "hrfi"),
            (lazy_fixture("FakeFCIFileHandlerFDHSIIQTI_fixture"), "hdfi"),
            (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture"), "hdfi"),
            (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"), "hrfi"),
                                                        ])
    def test_load_calibration(self, reader_configs, fh_param,
                              caplog, calibration, res_type):
        """Test loading with counts,radiance,reflectance and bt."""
        expected_res_n = {}
        if calibration == "reflectance":
            list_chan = fh_param["channels"]["solar"]
            list_grid = fh_param["channels"]["solar_grid_type"]
            expected_res_n["hdfi"] = 8
            expected_res_n["hrfi"] = 2
        elif calibration == "brightness_temperature":
            list_chan = fh_param["channels"]["terran"]
            list_grid = fh_param["channels"]["terran_grid_type"]
            expected_res_n["hdfi"] = 8
            expected_res_n["hrfi"] = 2
        else:
            list_chan = fh_param["channels"]["solar"] + fh_param["channels"]["terran"]
            list_grid = fh_param["channels"]["solar_grid_type"] + fh_param["channels"]["terran_grid_type"]
            expected_res_n["hdfi"] = 16
            expected_res_n["hrfi"] = 4
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        with caplog.at_level(logging.WARNING):
            res = reader.load(
                [make_dataid(name=name, calibration=calibration) for name in
                 list_chan], pad_data=False)
            assert caplog.text == ""
        assert expected_res_n[res_type] == len(res)
        for ch, grid_type in zip(list_chan,
                                 list_grid):
            self._shape_test(res, ch, grid_type, DICT_CALIBRATION[calibration])
            self._get_assert_load(res, ch, DICT_CALIBRATION[calibration],
                                  fh_param["filenames"][0])

    @pytest.mark.parametrize("fh_param", [lazy_fixture("FakeFCIFileHandlerFDHSI_fixture")])
    def test_load_calibration_negative_rad(self, reader_configs, fh_param):
        """Test calibrating negative radiances.

        See https://github.com/pytroll/satpy/issues/3009.
        """
        import satpy
        reader = _get_reader_with_filehandlers(fh_param["filenames"],
                                               reader_configs,
                                               clip_negative_radiances=True)
        did = make_dataid(name="ir_38", calibration="radiance")
        res = reader.load([did], pad_data=False)
        with satpy.config.set({"readers.clip_negative_radiances": True}):
            reader2 = _get_reader_with_filehandlers(fh_param["filenames"],
                                                    reader_configs)
            res2 = reader2.load([did], pad_data=False)
        numpy.testing.assert_array_equal(res["ir_38"][-1, :], 5)  # smallest positive radiance
        numpy.testing.assert_array_equal(res2["ir_38"][-1, :], 5)  # smallest positive radiance
        assert res["ir_38"].dtype == res2["ir_38"].dtype == np.dtype("float32")

    @pytest.mark.parametrize("fh_param", [lazy_fixture("FakeFCIFileHandlerFDHSI_fixture")])
    def test_drop_fill_value(self, reader_configs, fh_param):
        """Test that _FillValue attribute is dropped for floats.

        See https://github.com/pytroll/satpy/issues/3058.
        """
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        did = make_dataid(name="ir_105", calibration="brightness_temperature")
        res = reader.load([did], pad_data=False)
        assert "_FillValue" not in res["ir_105"].attrs

        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        did = make_dataid(name="vis_06", calibration="reflectance")
        res = reader.load([did], pad_data=False)
        assert "_FillValue" not in res["vis_06"].attrs

        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        did = make_dataid(name="vis_06", calibration="radiance")
        res = reader.load([did], pad_data=False)
        assert "_FillValue" not in res["vis_06"].attrs

        # but for counts it should still be there
        did = make_dataid(name="ir_105", calibration="counts")
        res = reader.load([did], pad_data=False)
        assert "_FillValue" in res["ir_105"].attrs

    @pytest.mark.parametrize(("calibration", "channel", "resolution"), [
        (calibration, channel, resolution)
        for calibration in ["counts", "radiance", "brightness_temperature", "reflectance"]
        for channel, resolution in generate_parameters(calibration)
    ])
    def test_load_calibration_af(self, FakeFCIFileHandlerAF_fixture, reader_configs, channel, calibration, caplog):
        """Test loading with counts,radiance,reflectance and bt for AF files."""
        expected_res_n = 1
        fh_param = FakeFCIFileHandlerAF_fixture
        type_ter = self._get_type_ter_AF(channel)
        with caplog.at_level(logging.WARNING):
            res = self._get_res_AF(channel, fh_param, calibration, reader_configs)
            assert caplog.text == ""
        assert expected_res_n == len(res)
        for ch, grid_type in zip(fh_param["channels"][type_ter],
                                 fh_param["channels"][f"{type_ter}_grid_type"]):
            self._shape_test(res, ch, grid_type, DICT_CALIBRATION[calibration])
            self._get_assert_load(res, ch, DICT_CALIBRATION[calibration],
                                  fh_param["filenames"][0])

    @pytest.mark.parametrize("fh_param", [(lazy_fixture("FakeFCIFileHandlerFDHSI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIQ4_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIQ4_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIIQTI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIIQTI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"))
                                          ])
    def test_orbital_parameters_attr(self, reader_configs, fh_param):
        """Test the orbital parameter attribute."""
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        res = reader.load(
            [make_dataid(name=name) for name in
             fh_param["channels"]["solar"] + fh_param["channels"]["terran"]], pad_data=False)

        for ch in fh_param["channels"]["solar"] + fh_param["channels"]["terran"]:
            assert res[ch].attrs["orbital_parameters"] == {
                "satellite_actual_longitude": np.mean(np.arange(6000)) if "IQTI" not in
                                                                          fh_param["filenames"][0] else 0.0,
                "satellite_actual_latitude": np.mean(np.arange(6000)) if "IQTI" not in
                                                                         fh_param["filenames"][0] else 0.0,
                "satellite_actual_altitude": np.mean(np.arange(6000)) if "IQTI" not in
                                                                         fh_param["filenames"][0] else 35786400.0,
                "satellite_nominal_longitude": 0.0,
                "satellite_nominal_latitude": 0,
                "satellite_nominal_altitude": 35786400.0,
                "projection_longitude": 0.0,
                "projection_latitude": 0,
                "projection_altitude": 35786400.0,
            }

    @pytest.mark.parametrize(("fh_param", "expected_pos_info"), [
        (lazy_fixture("FakeFCIFileHandlerFDHSI_fixture"), EXPECTED_POS_INFO_FOR_FILETYPE["fdhsi"]),
        (lazy_fixture("FakeFCIFileHandlerHRFI_fixture"), EXPECTED_POS_INFO_FOR_FILETYPE["hrfi"]),
        (lazy_fixture("FakeFCIFileHandlerHRFIQ4_fixture"), EXPECTED_POS_INFO_FOR_FILETYPE["hrfi"]),
        (lazy_fixture("FakeFCIFileHandlerFDHSIQ4_fixture"), EXPECTED_POS_INFO_FOR_FILETYPE["fdhsi"]),
        (lazy_fixture("FakeFCIFileHandlerHRFIIQTI_fixture"), EXPECTED_POS_INFO_FOR_FILETYPE["hrfi"]),
        (lazy_fixture("FakeFCIFileHandlerFDHSIIQTI_fixture"), EXPECTED_POS_INFO_FOR_FILETYPE["fdhsi"]),
        (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture"), EXPECTED_POS_INFO_FOR_FILETYPE["fdhsi"]),
        (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"), EXPECTED_POS_INFO_FOR_FILETYPE["hrfi"])
    ])
    def test_get_segment_position_info(self, reader_configs, fh_param, expected_pos_info):
        """Test the segment position info method."""
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        for filetype_handler in list(reader.file_handlers.values())[0]:
            segpos_info = filetype_handler.get_segment_position_info()
            assert segpos_info == expected_pos_info

    @pytest.mark.parametrize(("channel", "resolution"), generate_parameters("radiance"))
    def test_not_get_segment_info_called_af(self, FakeFCIFileHandlerAF_fixture, reader_configs, channel, resolution):
        """Test that checks that the get_segment_position_info has not been called for AF data."""
        with mock.patch("satpy.readers.fci_l1c_nc.FCIL1cNCFileHandler.get_segment_position_info") as gspi:
            fh_param = FakeFCIFileHandlerAF_fixture
            reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
            with contextlib.suppress(KeyError):
                # attempt to load the channel
                # If get_segment_position_info is called, the code will fail with a KeyError because of the mocking.
                # However, the point of the test is to check if the function has been called, not if the function
                # would work with this case, so the expected KeyError is suppressed, and we assert_not_called below.
                reader.load([channel])
            gspi.assert_not_called()

    @pytest.mark.parametrize("calibration", ["index_map", "pixel_quality"])
    @pytest.mark.parametrize(("fh_param", "expected_res_n"), [
        (lazy_fixture("FakeFCIFileHandlerFDHSI_fixture"), 16),
        (lazy_fixture("FakeFCIFileHandlerHRFI_fixture"), 4),
        (lazy_fixture("FakeFCIFileHandlerHRFIQ4_fixture"), 4),
        (lazy_fixture("FakeFCIFileHandlerFDHSIQ4_fixture"), 16),
        (lazy_fixture("FakeFCIFileHandlerHRFIIQTI_fixture"), 4),
        (lazy_fixture("FakeFCIFileHandlerFDHSIIQTI_fixture"), 16),
        (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture"), 16),
        (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"), 4)])
    def test_load_map_and_pixel(self, reader_configs, fh_param, expected_res_n, calibration):
        """Test loading of index_map and pixel_quality."""
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        res = reader.load(
            [f"{name}_{calibration}" for name in
             fh_param["channels"]["solar"] + fh_param["channels"]["terran"]], pad_data=False)
        assert expected_res_n == len(res)
        for ch, grid_type in zip(fh_param["channels"]["solar"] + fh_param["channels"]["terran"],
                                 fh_param["channels"]["solar_grid_type"] +
                                 fh_param["channels"]["terran_grid_type"]):
            assert res[f"{ch}_{calibration}"].shape == (GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["nrows"],
                                                        GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["ncols"])
            if calibration == "index_map":
                numpy.testing.assert_array_equal(res[f"{ch}_{calibration}"][1, 1], 110)
            elif calibration == "pixel_quality":
                numpy.testing.assert_array_equal(res[f"{ch}_{calibration}"][1, 1], 3)
                assert res[f"{ch}_{calibration}"].attrs["name"] == ch + "_pixel_quality"
            # Verify that attributes are propagated/inserted
            for attr_key in REQUIRED_ATTR_KEYS:
                assert attr_key in res[f"{ch}_{calibration}"].attrs.keys(), f"Missing attribute: {attr_key}"
            assert res[f"{ch}_{calibration}"].attrs["platform_name"] == "Meteosat-12"

    @pytest.mark.parametrize(("calibration", "channel", "resolution"), [
        (calibration, channel, resolution)
        for calibration in ["index_map", "pixel_quality"]
        for channel, resolution in generate_parameters(calibration)
    ])
    def test_load_map_and_pixel_af(self, FakeFCIFileHandlerAF_fixture, reader_configs, channel, calibration):
        """Test loading with of index_map and pixel_quality for AF files."""
        expected_res_n = 1
        fh_param = FakeFCIFileHandlerAF_fixture
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        type_ter = self._get_type_ter_AF(channel)
        res = reader.load([f"{name}_{calibration}"
                           for name in fh_param["channels"][type_ter]], pad_data=False)
        assert expected_res_n == len(res)
        for ch, grid_type in zip(fh_param["channels"][type_ter],
                                 fh_param["channels"][f"{type_ter}_grid_type"]):
            assert res[f"{ch}_{calibration}"].shape == (GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["nrows"],
                                                        GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["ncols"])
            if calibration == "index_map":
                numpy.testing.assert_array_equal(res[f"{ch}_{calibration}"][1, 1], 110)
            elif calibration == "pixel_quality":
                numpy.testing.assert_array_equal(res[f"{ch}_{calibration}"][1, 1], 3)
                assert res[f"{ch}_{calibration}"].attrs["name"] == ch + "_pixel_quality"
            # Verify that attributes are propagated/inserted
            for attr_key in REQUIRED_ATTR_KEYS:
                assert attr_key in res[f"{ch}_{calibration}"].attrs.keys(), f"Missing attribute: {attr_key}"
            assert res[f"{ch}_{calibration}"].attrs["platform_name"] == "Meteosat-12"

    @pytest.mark.parametrize("fh_param", [(lazy_fixture("FakeFCIFileHandlerFDHSI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIQ4_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIQ4_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIIQTI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIIQTI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"))])
    def test_load_aux_data(self, reader_configs, fh_param):
        """Test loading of auxiliary data."""
        from satpy.readers.fci_l1c_nc import AUX_DATA
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        res = reader.load([fh_param["channels"]["solar"][0] + "_" + key for key in AUX_DATA.keys()],
                          pad_data=False)
        grid_type = fh_param["channels"]["solar_grid_type"][0]
        for aux in [fh_param["channels"]["solar"][0] + "_" + key for key in AUX_DATA.keys()]:
            assert res[aux].shape == (GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["nrows"],
                                      GRID_TYPE_INFO_FOR_TEST_CONTENT[grid_type]["ncols"])
            if (aux == fh_param["channels"]["solar"][0] + "_earth_sun_distance") and ("IQTI" not in
                                                                                      fh_param["filenames"][0]):
                numpy.testing.assert_array_equal(res[aux][1, 1], 149597870.7)
            elif aux == fh_param["channels"]["solar"][0] + "_earth_sun_distance":
                numpy.testing.assert_array_equal(res[aux][1, 1], np.nan)
            else:
                numpy.testing.assert_array_equal(res[aux][1, 1], 10)
            # Verify that attributes are propagated/inserted
            assert res[aux].attrs["platform_name"] == "Meteosat-12"
            assert "time_parameters" in res[aux].attrs.keys()
            assert "orbital_parameters" in res[aux].attrs.keys()
            assert "file_type" in res[aux].attrs.keys()
            assert "name" in res[aux].attrs.keys()
            assert "sensor" in res[aux].attrs.keys()
            assert "resolution" in res[aux].attrs.keys()

    @pytest.mark.parametrize("fh_param", [(lazy_fixture("FakeFCIFileHandlerFDHSI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIQ4_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIQ4_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIIQTI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIIQTI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"))
                                          ])
    def test_platform_name(self, reader_configs, fh_param):
        """Test that platform name is exposed.

        Test that the FCI reader exposes the platform name.  Corresponds
        to GH issue 1014.
        """
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        res = reader.load(["vis_06"], pad_data=False)
        assert res["vis_06"].attrs["platform_name"] == "Meteosat-12"

    @pytest.mark.parametrize(("fh_param", "compare_tuples"),
                             [(lazy_fixture("FakeFCIFileHandlerFDHSI_fixture"), (67, 10,
                                                                                 datetime.datetime.strptime(
                                                                                     "2017-04-10 11:30:00",
                                                                                     "%Y-%m-%d %H:%M:%S"),
                                                                                 datetime.datetime.strptime(
                                                                                     "2017-04-10 11:40:00",
                                                                                     "%Y-%m-%d %H:%M:%S"))),
                              (lazy_fixture("FakeFCIFileHandlerHRFI_fixture"), (67, 10,
                                                                                datetime.datetime.strptime(
                                                                                    "2017-04-10 11:30:00",
                                                                                    "%Y-%m-%d %H:%M:%S"),
                                                                                datetime.datetime.strptime(
                                                                                    "2017-04-10 11:40:00",
                                                                                    "%Y-%m-%d %H:%M:%S"))),
                              (lazy_fixture("FakeFCIFileHandlerHRFIQ4_fixture"), (29, 2.5,
                                                                                  datetime.datetime.strptime(
                                                                                      "2023-07-22 12:00:00",
                                                                                      "%Y-%m-%d %H:%M:%S"),
                                                                                  datetime.datetime.strptime(
                                                                                      "2023-07-22 12:02:30",
                                                                                      "%Y-%m-%d %H:%M:%S"))),
                              (lazy_fixture("FakeFCIFileHandlerFDHSIQ4_fixture"), (29, 2.5,
                                                                                   datetime.datetime.strptime(
                                                                                       "2023-07-22 12:00:00",
                                                                                       "%Y-%m-%d %H:%M:%S"),
                                                                                   datetime.datetime.strptime(
                                                                                       "2023-07-22 12:02:30",
                                                                                       "%Y-%m-%d %H:%M:%S"))),
                              (lazy_fixture("FakeFCIFileHandlerHRFIIQTI_fixture"), (1, 10,
                                                                                    datetime.datetime.strptime(
                                                                                        "2023-10-16 12:50:00",
                                                                                        "%Y-%m-%d %H:%M:%S"),
                                                                                    datetime.datetime.strptime(
                                                                                        "2023-10-16 13:00:00",
                                                                                        "%Y-%m-%d %H:%M:%S"))),
                              (lazy_fixture("FakeFCIFileHandlerFDHSIIQTI_fixture"), (1, 10,
                                                                                    datetime.datetime.strptime(
                                                                                         "2023-10-16 12:50:00",
                                                                                         "%Y-%m-%d %H:%M:%S"),
                                                                                    datetime.datetime.strptime(
                                                                                         "2023-10-16 13:00:00",
                                                                                         "%Y-%m-%d %H:%M:%S"))),
                              (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture"), (2, None,
                                                                                    datetime.datetime.strptime(
                                                                                         "2023-07-22 12:00:00",
                                                                                         "%Y-%m-%d %H:%M:%S"),
                                                                                    datetime.datetime.strptime(
                                                                                         "2023-07-22 12:00:27",
                                                                                         "%Y-%m-%d %H:%M:%S"))),
                              (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"), (2, None,
                                                                                    datetime.datetime.strptime(
                                                                                         "2023-07-22 12:00:00",
                                                                                         "%Y-%m-%d %H:%M:%S"),
                                                                                    datetime.datetime.strptime(
                                                                                         "2023-07-22 12:00:27",
                                                                                         "%Y-%m-%d %H:%M:%S")))
                              ])
    def test_count_in_repeat_cycle_rc_period_min(self, reader_configs, fh_param, compare_tuples):
        """Test the rc_period_min value for each configuration."""
        self._compare_rc_period_min_count_in_repeat_cycle(fh_param["filetype"], fh_param,
                                                          reader_configs, compare_tuples)

    @pytest.mark.parametrize(("channel", "resolution", "compare_tuples"),
                             [("vis_06", "3km", (1, 10,
                                                 datetime.datetime.strptime("2024-01-09 08:00:00", "%Y-%m-%d %H:%M:%S"),
                                                 datetime.datetime.strptime("2024-01-09 08:10:00",
                                                                            "%Y-%m-%d %H:%M:%S"))),
                              ("vis_06", "1km", (1, 10,
                                                 datetime.datetime.strptime("2024-01-09 08:00:00", "%Y-%m-%d %H:%M:%S"),
                                                 datetime.datetime.strptime("2024-01-09 08:10:00",
                                                                            "%Y-%m-%d %H:%M:%S")))
                              ])
    def test_count_in_repeat_cycle_rc_period_min_AF(self, FakeFCIFileHandlerAF_fixture, reader_configs,
                                                    channel, resolution, compare_tuples):
        """Test the rc_period_min value for each configuration."""
        fh_param = FakeFCIFileHandlerAF_fixture
        self._compare_rc_period_min_count_in_repeat_cycle(f"{fh_param['filetype']}_{channel}_{resolution}", fh_param,
                                                          reader_configs, compare_tuples)

    @pytest.mark.parametrize(("fh_param"),
                             [(lazy_fixture("FakeFCIFileHandlerFDHSI_fixture")),
                              (lazy_fixture("FakeFCIFileHandlerHRFI_fixture")),
                              (lazy_fixture("FakeFCIFileHandlerHRFIQ4_fixture")),
                              (lazy_fixture("FakeFCIFileHandlerFDHSIQ4_fixture")),
                              (lazy_fixture("FakeFCIFileHandlerHRFIIQTI_fixture")),
                              (lazy_fixture("FakeFCIFileHandlerFDHSIIQTI_fixture")),
                              (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture")),
                              (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"))
                              ])
    def test_compute_earth_sun_parameter(self, reader_configs, fh_param):
        """Test the computation of the sun_earth_parameter."""
        self._compare_sun_earth_distance(fh_param["filetype"], fh_param, reader_configs)

    @pytest.mark.parametrize(("channel", "resolution"), [("vis_06", "3km"),
                                                         ("vis_06", "1km")])
    def test_compute_earth_sun_parameter_AF(self, FakeFCIFileHandlerAF_fixture, reader_configs,
                                            channel, resolution):
        """Test the rc_period_min value for each configuration."""
        fh_param = FakeFCIFileHandlerAF_fixture
        self._compare_sun_earth_distance(f"{fh_param['filetype']}_{channel}_{resolution}", fh_param, reader_configs)

    @pytest.mark.parametrize(("fh_param"), [ (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture")),
                                             (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"))])
    def test_rc_period_min_log_message_other_coverage(self, reader_configs, fh_param, caplog):
        """Test the rc_period_min error."""
        with caplog.at_level(logging.DEBUG):
            _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        assert any('Coverage "xx" not recognised. Using observation times for nominal times.'
                   in message for message in caplog.messages)

    @pytest.mark.parametrize(("fh_param", "expected_area"), [
        (lazy_fixture("FakeFCIFileHandlerFDHSI_fixture"), ["mtg_fci_fdss_1km", "mtg_fci_fdss_2km"]),
        (lazy_fixture("FakeFCIFileHandlerHRFI_fixture"), ["mtg_fci_fdss_500m", "mtg_fci_fdss_1km"]),
        (lazy_fixture("FakeFCIFileHandlerHRFIQ4_fixture"), ["mtg_fci_fdss_500m", "mtg_fci_fdss_1km"]),
        (lazy_fixture("FakeFCIFileHandlerFDHSIQ4_fixture"), ["mtg_fci_fdss_1km", "mtg_fci_fdss_2km"]),
        (lazy_fixture("FakeFCIFileHandlerHRFIIQTI_fixture"), ["mtg_fci_fdss_500m", "mtg_fci_fdss_1km"]),
        (lazy_fixture("FakeFCIFileHandlerFDHSIIQTI_fixture"), ["mtg_fci_fdss_1km", "mtg_fci_fdss_2km"]),
        (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture"), ["mtg_fci_fdss_1km", "mtg_fci_fdss_2km"]),
        (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"), ["mtg_fci_fdss_500m", "mtg_fci_fdss_1km"])
    ])
    def test_area_definition_computation(self, reader_configs, fh_param, expected_area):
        """Test that the geolocation computation is correct."""
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)
        res = reader.load(["ir_105", "vis_06"], pad_data=False)

        # test that area_ids are harmonisation-conform <platform>_<instrument>_<service>_<resolution>
        assert res["vis_06"].attrs["area"].area_id == expected_area[0]
        assert res["ir_105"].attrs["area"].area_id == expected_area[1]

        area_def = res["ir_105"].attrs["area"]
        # test area extents computation
        np.testing.assert_array_almost_equal(np.array(area_def.area_extent),
                                             np.array([-5567999.994203, -5367999.994411,
                                                       5567999.994203, -5567999.994203]),
                                             decimal=2)

        # check that the projection is read in properly
        assert area_def.crs.coordinate_operation.method_name == "Geostationary Satellite (Sweep Y)"
        assert area_def.crs.coordinate_operation.params[0].value == 0.0  # projection origin longitude
        assert area_def.crs.coordinate_operation.params[1].value == 35786400.0  # projection height
        assert area_def.crs.ellipsoid.semi_major_metre == 6378137.0
        assert area_def.crs.ellipsoid.inverse_flattening == 298.257223563
        assert area_def.crs.ellipsoid.is_semi_minor_computed

    @pytest.mark.parametrize("fh_param", [(lazy_fixture("FakeFCIFileHandlerFDHSI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIQ4_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIQ4_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIIQTI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIIQTI_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerFDHSIOtherCoverage_fixture")),
                                          (lazy_fixture("FakeFCIFileHandlerHRFIOtherCoverage_fixture"))
                                          ])
    def test_excs(self, reader_configs, fh_param):
        """Test that exceptions are raised where expected."""
        reader = _get_reader_with_filehandlers(fh_param["filenames"], reader_configs)

        with pytest.raises(ValueError, match="Unknown dataset key, not a channel, quality or auxiliary data: invalid"):
            reader.file_handlers[fh_param["filetype"]][0].get_dataset(make_dataid(name="invalid"), {})
        with pytest.raises(ValueError, match="unknown invalid value for <enum 'calibration'>"):
            reader.file_handlers[fh_param["filetype"]][0].get_dataset(
                make_dataid(name="ir_123", calibration="unknown"),
                {"units": "unknown"})

    def test_load_composite(self):
        """Test that composites are loadable."""
        # when dedicated composites for FCI are implemented in satpy,
        # this method should probably move to a dedicated class and module
        # in the tests.compositor_tests package

        from satpy.composites.config_loader import load_compositor_configs_for_sensors
        comps, mods = load_compositor_configs_for_sensors(["fci"])
        assert len(comps["fci"]) > 0
        assert len(mods["fci"]) > 0


class TestFCIL1cNCReaderBadData:
    """Test the FCI L1c NetCDF Reader for bad data input."""

    def test_handling_bad_data_ir(self, reader_configs, caplog):
        """Test handling of bad IR data."""
        with mocked_basefilehandler(FakeFCIFileHandlerWithBadData):
            reader = _get_reader_with_filehandlers(TEST_FILENAMES["fdhsi"], reader_configs)
            with caplog.at_level(logging.ERROR):
                reader.load([make_dataid(
                    name="ir_105",
                    calibration="brightness_temperature")], pad_data=False)
                assert "cannot produce brightness temperature" in caplog.text

    def test_handling_bad_data_vis(self, reader_configs, caplog):
        """Test handling of bad VIS data."""
        with mocked_basefilehandler(FakeFCIFileHandlerWithBadData):
            reader = _get_reader_with_filehandlers(TEST_FILENAMES["fdhsi"], reader_configs)
            with caplog.at_level(logging.ERROR):
                reader.load([make_dataid(
                    name="vis_06",
                    calibration="reflectance")], pad_data=False)
                assert "cannot produce reflectance" in caplog.text



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018, 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""MITIFF writer objects for creating MITIFF files from `Dataset` objects."""
from __future__ import annotations

import logging
import os
import typing

import dask
import numpy as np
from PIL import Image, ImagePalette

from satpy.dataset import DataID, DataQuery
from satpy.enhancements.enhancer import get_enhanced_image
from satpy.writers.core.image import ImageWriter

if typing.TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any

    import dask.array as da
    import xarray as xr
    from dask.delayed import Delayed

IMAGEDESCRIPTION = 270

LOG = logging.getLogger(__name__)

KELVIN_TO_CELSIUS = -273.15


def _adjust_kwargs(dataset, kwargs):
    if "platform_name" not in kwargs:
        kwargs["platform_name"] = dataset.attrs["platform_name"]
    if "name" not in kwargs:
        kwargs["name"] = dataset.attrs["name"]
    if "start_time" not in kwargs:
        kwargs["start_time"] = dataset.attrs["start_time"]
    if "sensor" not in kwargs:
        kwargs["sensor"] = dataset.attrs["sensor"]
    # Sensor attrs could be set. MITIFFs needing to handle sensor can only have one sensor
    # Assume the first value of set as the sensor.
    if isinstance(kwargs["sensor"], set):
        LOG.warning("Sensor is set, will use the first value: %s", kwargs["sensor"])
        kwargs["sensor"] = (list(kwargs["sensor"]))[0]


class MITIFFWriter(ImageWriter):
    """Writer to produce MITIFF image files."""

    def __init__(self, name=None, tags=None, **kwargs):
        """Initialize reader with tag and other configuration information."""
        super().__init__(name=name, default_config_filename="writers/mitiff.yaml", **kwargs)
        self.tags = self.info.get("tags", None) if tags is None else tags
        if self.tags is None:
            self.tags = {}
        elif not isinstance(self.tags, dict):
            # if it's coming from a config file
            self.tags = dict(tuple(x.split("=")) for x in self.tags.split(","))

        self.mitiff_config = {}
        self.translate_channel_name = {}
        self.channel_order = {}
        self.palette = False
        self.sensor = None

    def save_dataset(
            self,
            dataset: xr.DataArray,
            filename: str | None = None,
            fill_value: float | int | None = None,
            compute: bool = True,
            units: str | None = None,
            overlay: dict | None = None,
            decorate: dict | None = None,
            **kwargs,
        ) -> list[da.Array | Delayed] | tuple[list[da.Array], list[Any]] | list[str | os.PathLike | None]:
        """Save single dataset as mitiff file."""
        LOG.debug("Starting in mitiff save_dataset ... ")

        if fill_value is not None:
            raise ValueError("'fill_value' is not used by the MITIFF writer.")
        if overlay is not None or decorate is not None:
            raise ValueError("'overlay' and 'decorate' options are not implemented by the MITIFF writer.")
        # to match parent class functionality and interface:
        if units is not None:
            import pint_xarray  # noqa
            dataset = dataset.pint.quantify().pint.to(units).pint.dequantify()

        def _delayed_create(dataset):
            try:
                if "palette" in kwargs:
                    self.palette = kwargs["palette"]
                _adjust_kwargs(dataset, kwargs)

                try:
                    self.mitiff_config[kwargs["sensor"]] = dataset.attrs["metadata_requirements"]["config"]
                    self.channel_order[kwargs["sensor"]] = dataset.attrs["metadata_requirements"]["order"]
                    self.file_pattern = dataset.attrs["metadata_requirements"]["file_pattern"]
                except KeyError:
                    # For some mitiff products this info is needed, for others not.
                    # If needed you should know how to fix this
                    pass

                try:
                    self.translate_channel_name[kwargs["sensor"]] = \
                        dataset.attrs["metadata_requirements"]["translate"]
                except KeyError:
                    # For some mitiff products this info is needed, for others not.
                    # If needed you should know how to fix this
                    pass

                image_description = self._make_image_description(dataset, **kwargs)
                gen_filename = filename or self.get_filename(**dataset.attrs)
                LOG.info("Saving mitiff to: %s ...", gen_filename)
                self._save_datasets_as_mitiff(dataset, image_description,
                                              gen_filename, **kwargs)
            except (KeyError, ValueError, RuntimeError):
                raise

            return gen_filename

        delayed = dask.delayed(_delayed_create)(dataset)

        if compute:
            return [delayed.compute()]
        return [delayed]

    def save_datasets(
            self,
            datasets: Iterable[xr.DataArray],
            compute: bool = True,
            filename: str | None = None,
            **kwargs,
    ) -> list[da.Array | Delayed] | tuple[list[da.Array], list[Any]] | list[str | os.PathLike | None]:
        """Save all datasets to one or more files."""
        LOG.debug("Starting in mitiff save_datasets ... ")

        def _delayed_create(datasets):
            dataset = datasets[0]

            try:
                _adjust_kwargs(dataset, kwargs)

                try:
                    self.mitiff_config[kwargs["sensor"]] = dataset.attrs["metadata_requirements"]["config"]
                    translate = dataset.attrs["metadata_requirements"]["translate"]
                    self.translate_channel_name[kwargs["sensor"]] = translate
                    self.channel_order[kwargs["sensor"]] = dataset.attrs["metadata_requirements"]["order"]
                    self.file_pattern = dataset.attrs["metadata_requirements"]["file_pattern"]
                except KeyError:
                    # For some mitiff products this info is needed, for others not.
                    # If needed you should know how to fix this
                    pass

                image_description = self._make_image_description(datasets, **kwargs)
                LOG.debug("File pattern %s", self.file_pattern)
                if isinstance(datasets, list):
                    kwargs["start_time"] = dataset.attrs["start_time"]
                else:
                    kwargs["start_time"] = datasets.attrs["start_time"]
                gen_filename = filename or self.get_filename(**kwargs)
                LOG.info("Saving mitiff to: %s ...", gen_filename)
                self._save_datasets_as_mitiff(datasets, image_description, gen_filename, **kwargs)
            except (KeyError, ValueError, RuntimeError):
                raise

            return gen_filename

        delayed = dask.delayed(_delayed_create)(datasets)
        LOG.debug("About to call delayed compute ...")
        if compute:
            return [delayed.compute()]
        return [delayed]

    def _make_channel_list(self, datasets, **kwargs):
        channels = []
        try:
            if self.channel_order:
                channels = self._reorder_channels(datasets, **kwargs)
            elif self.palette:
                if "palette_channel_name" in kwargs:
                    channels.append(kwargs["palette_channel_name"].upper())
                else:
                    LOG.error("Is palette but can not find palette_channel_name to name the dataset")
            else:
                for ch in range(len(datasets)):
                    channels.append(ch + 1)
        except KeyError:
            for ch in range(len(datasets)):
                channels.append(ch + 1)
        return channels

    def _reorder_channels(self, datasets, **kwargs):
        channels = []
        for cn in self.channel_order[kwargs["sensor"]]:
            for ch, ds in enumerate(datasets):
                if isinstance(ds.attrs["prerequisites"][ch], (DataQuery, DataID)):
                    if ds.attrs["prerequisites"][ch]["name"] == cn:
                        channels.append(
                            ds.attrs["prerequisites"][ch]["name"])
                        break
                else:
                    if ds.attrs["prerequisites"][ch] == cn:
                        channels.append(
                            ds.attrs["prerequisites"][ch])
                        break
        return channels

    def _channel_names(self, channels, cns, **kwargs):
        _image_description = ""
        for ch in channels:
            try:
                _image_description += str(
                    self.mitiff_config[kwargs["sensor"]][cns.get(ch, ch)]["alias"])
            except KeyError:
                _image_description += str(ch)
            _image_description += " "

        # Replace last char(space) with \n
        _image_description = _image_description[:-1]
        _image_description += "\n"

        return _image_description

    def _add_sizes(self, datasets, first_dataset):
        _image_description = " Xsize: "
        if isinstance(datasets, list):
            _image_description += str(first_dataset.sizes["x"]) + "\n"
        else:
            _image_description += str(datasets.sizes["x"]) + "\n"

        _image_description += " Ysize: "
        if isinstance(datasets, list):
            _image_description += str(first_dataset.sizes["y"]) + "\n"
        else:
            _image_description += str(datasets.sizes["y"]) + "\n"

        return _image_description

    def _add_proj4_string(self, datasets, first_dataset, **kwargs):
        import warnings

        proj4_string = " Proj string: "

        if isinstance(datasets, list):
            area = first_dataset.attrs["area"]
        else:
            area = datasets.attrs["area"]
        # Use pyproj's CRS object to get a valid EPSG code if possible
        # only in newer pyresample versions with pyproj 2.0+ installed
        if hasattr(area, "crs") and area.crs.to_epsg() is not None:
            proj4_string += "+init=EPSG:{}".format(area.crs.to_epsg())
        else:
            # Filter out the PROJ warning of losing projection information
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=UserWarning,
                                        message=r"You will likely lose important projection information")
                proj4_string += area.proj_str

        x_0 = 0
        y_0 = 0
        # FUTURE: Use pyproj 2.0+ to convert EPSG to PROJ4 if possible
        proj4_string, x_0 = self._convert_epsg_to_proj(proj4_string, x_0)

        proj4_string = self._special_correction_of_proj_string(proj4_string)

        if isinstance(datasets, list):
            _dataset = first_dataset
        else:
            _dataset = datasets
        mitiff_pixel_adjustment = kwargs.get("mitiff_pixel_adjustment", True)
        proj4_string = self._append_projection_center(proj4_string, _dataset, x_0, y_0, mitiff_pixel_adjustment)
        LOG.debug("proj4_string: %s", proj4_string)
        proj4_string += "\n"

        return proj4_string

    def _special_correction_of_proj_string(self, proj4_string):
        if "geos" in proj4_string:
            proj4_string = proj4_string.replace("+sweep=x ", "")
            if "+a=6378137.0 +b=6356752.31414" in proj4_string:
                proj4_string = proj4_string.replace("+a=6378137.0 +b=6356752.31414",
                                                    "+ellps=WGS84")
        if "+units=m" in proj4_string:
            proj4_string = proj4_string.replace("+units=m", "+units=km")

        if not any(datum in proj4_string for datum in ["datum", "towgs84"]):
            proj4_string += " +towgs84=0,0,0"

        if "units" not in proj4_string:
            proj4_string += " +units=km"
        return proj4_string

    def _append_projection_center(self, proj4_string, dataset, x_0, y_0, mitiff_pixel_adjustment):
        corner_correction_x, corner_correction_y = self._set_correction_size(dataset, mitiff_pixel_adjustment)
        if "x_0" not in proj4_string:
            proj4_string += " +x_0=%.6f" % (
                    (-dataset.attrs["area"].area_extent[0] +
                     corner_correction_x) + x_0)
            proj4_string += " +y_0=%.6f" % (
                    (-dataset.attrs["area"].area_extent[1] +
                     corner_correction_y) + y_0)
        elif "+x_0=0" in proj4_string and "+y_0=0" in proj4_string:
            proj4_string = proj4_string.replace("+x_0=0", "+x_0=%.6f" % (
                    (-dataset.attrs["area"].area_extent[0] +
                     corner_correction_x) + x_0))
            proj4_string = proj4_string.replace("+y_0=0", "+y_0=%.6f" % (
                    (-dataset.attrs["area"].area_extent[1] +
                     corner_correction_y) + y_0))
        return proj4_string

    def _set_correction_size(self, dataset, mitiff_pixel_adjustment):
        corner_correction_x = dataset.attrs["area"].pixel_size_x
        corner_correction_y = dataset.attrs["area"].pixel_size_y
        if not mitiff_pixel_adjustment:
            corner_correction_x = 0
            corner_correction_y = 0
        return corner_correction_x,corner_correction_y

    def _convert_epsg_to_proj(self, proj4_string, x_0):
        if "EPSG:32631" in proj4_string:
            proj4_string = proj4_string.replace("+init=EPSG:32631",
                                                "+proj=etmerc +lat_0=0 +lon_0=3 +k=0.9996 +ellps=WGS84 +datum=WGS84")
            x_0 = 500000
        elif "EPSG:32632" in proj4_string:
            proj4_string = proj4_string.replace("+init=EPSG:32632",
                                                "+proj=etmerc +lat_0=0 +lon_0=9 +k=0.9996 +ellps=WGS84 +datum=WGS84")
            x_0 = 500000
        elif "EPSG:32633" in proj4_string:
            proj4_string = proj4_string.replace("+init=EPSG:32633",
                                                "+proj=etmerc +lat_0=0 +lon_0=15 +k=0.9996 +ellps=WGS84 +datum=WGS84")
            x_0 = 500000
        elif "EPSG:32634" in proj4_string:
            proj4_string = proj4_string.replace("+init=EPSG:32634",
                                                "+proj=etmerc +lat_0=0 +lon_0=21 +k=0.9996 +ellps=WGS84 +datum=WGS84")
            x_0 = 500000
        elif "EPSG:32635" in proj4_string:
            proj4_string = proj4_string.replace("+init=EPSG:32635",
                                                "+proj=etmerc +lat_0=0 +lon_0=27 +k=0.9996 +ellps=WGS84 +datum=WGS84")
            x_0 = 500000
        elif "EPSG" in proj4_string:
            LOG.warning("EPSG used in proj string but not converted. Please add this in code")
        return proj4_string, x_0

    def _add_pixel_sizes(self, datasets, first_dataset):
        _image_description = ""
        if isinstance(datasets, list):
            _image_description += " Ax: %.6f" % (
                first_dataset.attrs["area"].pixel_size_x / 1000.)
            _image_description += " Ay: %.6f" % (
                first_dataset.attrs["area"].pixel_size_y / 1000.)
        else:
            _image_description += " Ax: %.6f" % (
                datasets.attrs["area"].pixel_size_x / 1000.)
            _image_description += " Ay: %.6f" % (
                datasets.attrs["area"].pixel_size_y / 1000.)

        return _image_description

    def _add_corners(self, datasets, first_dataset):
        # But this ads up to upper left corner of upper left pixel.
        # But need to use the center of the pixel.
        # Therefor use the center of the upper left pixel.
        _image_description = ""
        if isinstance(datasets, list):
            _image_description += " Bx: %.6f" % (
                first_dataset.attrs["area"].area_extent[0] / 1000. +
                first_dataset.attrs["area"].pixel_size_x / 1000. / 2.)  # LL_x
            _image_description += " By: %.6f" % (
                first_dataset.attrs["area"].area_extent[3] / 1000. -
                first_dataset.attrs["area"].pixel_size_y / 1000. / 2.)  # UR_y
        else:
            _image_description += " Bx: %.6f" % (
                datasets.attrs["area"].area_extent[0] / 1000. +
                datasets.attrs["area"].pixel_size_x / 1000. / 2.)  # LL_x
            _image_description += " By: %.6f" % (
                datasets.attrs["area"].area_extent[3] / 1000. -
                datasets.attrs["area"].pixel_size_y / 1000. / 2.)  # UR_y

        _image_description += "\n"
        return _image_description

    def _add_calibration_datasets(self, ch, datasets, reverse_offset, reverse_scale, decimals):
        _reverse_offset = reverse_offset
        _reverse_scale = reverse_scale
        _decimals = decimals
        _table_calibration = ""
        found_calibration = False
        skip_calibration = False
        ds_list = datasets
        if not isinstance(datasets, list) and "bands" not in datasets.sizes:
            ds_list = [datasets]

        for i, ds in enumerate(ds_list):
            if ("prerequisites" in ds.attrs and
                isinstance(ds.attrs["prerequisites"], list) and
                len(ds.attrs["prerequisites"]) >= i + 1 and
                    isinstance(ds.attrs["prerequisites"][i], (DataQuery, DataID))):
                if ds.attrs["prerequisites"][i].get("name") == str(ch):
                    if ds.attrs["prerequisites"][i].get("calibration") == "RADIANCE":
                        raise NotImplementedError(
                            "Mitiff radiance calibration not implemented.")
                    # _table_calibration += ', Radiance, '
                    # _table_calibration += '[W/m²/µm/sr]'
                    # _decimals = 8
                    elif ds.attrs["prerequisites"][i].get("calibration") == "brightness_temperature":
                        found_calibration = True
                        _table_calibration += ", BT, "
                        _table_calibration += "\N{DEGREE SIGN}"
                        _table_calibration += u"[C]"

                        _reverse_offset = 255.
                        _reverse_scale = -1.
                        _decimals = 2
                    elif ds.attrs["prerequisites"][i].get("calibration") == "reflectance":
                        found_calibration = True
                        _table_calibration += ", Reflectance(Albedo), "
                        _table_calibration += "[%]"
                        _decimals = 2
                    else:
                        LOG.warning("Unknown calib type. Must be Radiance, Reflectance or BT.")

                        break
                else:
                    continue
            else:
                _table_calibration = ""
                skip_calibration = True
                break
        if not found_calibration:
            _table_calibration = ""
            skip_calibration = True
            # How to format string by passing the format
            # http://stackoverflow.com/questions/1598579/rounding-decimals-with-new-python-format-function
        return skip_calibration, _table_calibration, _reverse_offset, _reverse_scale, _decimals

    def _add_palette_info(self, datasets, palette_unit, palette_description, **kwargs):
        # mitiff key word for palette interpretion
        _palette = "\n COLOR INFO:\n"
        # mitiff info for the unit of the interpretion
        _palette += " {}\n".format(palette_unit)
        # The length of the palette description as needed by mitiff in DIANA
        _palette += " {}\n".format(len(palette_description))
        for desc in palette_description:
            _palette += " {}\n".format(desc)
        return _palette

    def _add_calibration(self, channels, cns, datasets, **kwargs):

        _table_calibration = ""
        skip_calibration = False
        for ch in channels:

            palette = False
            # Make calibration.
            if palette:
                raise NotImplementedError("Mitiff palette saving is not implemented.")
            else:
                _table_calibration += "Table_calibration: "
                try:
                    _table_calibration += str(
                        self.mitiff_config[kwargs["sensor"]][cns.get(ch, ch)]["alias"])
                except KeyError:
                    _table_calibration += str(ch)

                _reverse_offset = 0.
                _reverse_scale = 1.
                _decimals = 2

                skip_calibration, __table_calibration, _reverse_offset, _reverse_scale, _decimals = \
                    self._add_calibration_datasets(ch, datasets, _reverse_offset, _reverse_scale, _decimals)
                _table_calibration += __table_calibration

                if not skip_calibration:
                    _table_calibration += ", 8, [ "
                    for val in range(0, 256):
                        # Comma separated list of values
                        _table_calibration += "{0:.{1}f} ".format((float(self.mitiff_config[
                            kwargs["sensor"]][cns.get(ch, ch)]["min-val"]) +
                            ((_reverse_offset + _reverse_scale * val) *
                             (float(self.mitiff_config[kwargs["sensor"]][cns.get(ch, ch)]["max-val"]) -
                             float(self.mitiff_config[kwargs["sensor"]][cns.get(ch, ch)]["min-val"]))) / 255.),
                            _decimals)
                        # _table_calibration += '0.00000000 '

                    _table_calibration += "]\n\n"
                else:
                    _table_calibration = ""

        return _table_calibration

    def _make_image_description(self, datasets, **kwargs):
        r"""Generate image description for mitiff.

        Satellite: NOAA 18
        Date and Time: 06:58 31/05-2016
        SatDir: 0
        Channels:   6 In this file: 1-VIS0.63 2-VIS0.86 3(3B)-IR3.7
        4-IR10.8 5-IR11.5 6(3A)-VIS1.6
        Xsize:  4720
        Ysize:  5544
        Map projection: Stereographic
        Proj string: +proj=stere +lon_0=0 +lat_0=90 +lat_ts=60
        +ellps=WGS84 +towgs84=0,0,0 +units=km
        +x_0=2526000.000000 +y_0=5806000.000000
        TrueLat: 60 N
        GridRot: 0
        Xunit:1000 m Yunit: 1000 m
        NPX: 0.000000 NPY: 0.000000
        Ax: 1.000000 Ay: 1.000000 Bx: -2526.000000 By: -262.000000

        Satellite: <satellite name>
        Date and Time: <HH:MM dd/mm-yyyy>
        SatDir: 0
        Channels:   <number of chanels> In this file: <channels names in order>
        Xsize:  <number of pixels x>
        Ysize:  <number of pixels y>
        Map projection: Stereographic
        Proj string: <proj4 string with +x_0 and +y_0 which is the positive
        distance from proj origo
        to the lower left corner of the image data>
        TrueLat: 60 N
        GridRot: 0
        Xunit:1000 m Yunit: 1000 m
        NPX: 0.000000 NPY: 0.000000
        Ax: <pixels size x in km> Ay: <pixel size y in km> Bx: <left corner of
        upper right pixel in km>
        By: <upper corner of upper right pixel in km>


        if palette image write special palette
        if normal channel write table calibration:
        Table_calibration: <channel name>, <calibration type>, [<unit>],
        <no of bits of data>,
        [<calibration values space separated>]\n\n

        """
        translate_platform_name = {"metop01": "Metop-B",
                                   "metop02": "Metop-A",
                                   "metop03": "Metop-C",
                                   "noaa15": "NOAA-15",
                                   "noaa16": "NOAA-16",
                                   "noaa17": "NOAA-17",
                                   "noaa18": "NOAA-18",
                                   "noaa19": "NOAA-19"}

        first_dataset = datasets
        if isinstance(datasets, list):
            LOG.debug("Datasets is a list of dataset")
            first_dataset = datasets[0]

        _platform_name = self._get_platform_name(first_dataset, translate_platform_name, kwargs)

        _image_description = ""
        _image_description.encode("utf-8")

        _image_description += " Satellite: "
        if _platform_name is not None:
            _image_description += _platform_name

        _image_description += "\n"

        _image_description += " Date and Time: "
        # Select earliest start_time
        first = True
        earliest = 0
        for dataset in datasets:
            if first:
                earliest = dataset.attrs["start_time"]
            else:
                if dataset.attrs["start_time"] < earliest:
                    earliest = dataset.attrs["start_time"]
            first = False
        LOG.debug("earliest start_time: %s", earliest)
        _image_description += earliest.strftime("%H:%M %d/%m-%Y\n")

        _image_description += " SatDir: 0\n"

        _image_description += " Channels: "

        _image_description += self._get_dataset_len(datasets)
        _image_description += " In this file: "

        channels = self._make_channel_list(datasets, **kwargs)

        try:
            cns = self.translate_channel_name.get(kwargs["sensor"], {})
        except KeyError:
            pass

        _image_description += self._channel_names(channels, cns, **kwargs)

        _image_description += self._add_sizes(datasets, first_dataset)

        _image_description += " Map projection: Stereographic\n"

        _image_description += self._add_proj4_string(datasets, first_dataset, **kwargs)

        _image_description += " TrueLat: 60N\n"
        _image_description += " GridRot: 0\n"

        _image_description += " Xunit:1000 m Yunit: 1000 m\n"

        _image_description += " NPX: %.6f" % (0)
        _image_description += " NPY: %.6f" % (0) + "\n"

        _image_description += self._add_pixel_sizes(datasets, first_dataset)
        _image_description += self._add_corners(datasets, first_dataset)

        if isinstance(datasets, list):
            LOG.debug("Area extent: %s", first_dataset.attrs["area"].area_extent)
        else:
            LOG.debug("Area extent: %s", datasets.attrs["area"].area_extent)

        if self.palette:
            LOG.debug("Doing palette image")
            _image_description += self._add_palette_info(datasets, **kwargs)
        else:
            _image_description += self._add_calibration(channels, cns, datasets, **kwargs)

        return _image_description

    def _get_dataset_len(self, datasets):
        if isinstance(datasets, list):
            LOG.debug("len datasets: %s", len(datasets))
            dataset_len = str(len(datasets))
        elif "bands" in datasets.sizes:
            LOG.debug("len datasets: %s", datasets.sizes["bands"])
            dataset_len = str(datasets.sizes["bands"])
        elif len(datasets.sizes) == 2:
            LOG.debug("len datasets: 1")
            dataset_len = "1"
        else:
            dataset_len = ""
        return dataset_len

    def _get_platform_name(self, first_dataset, translate_platform_name, kwargs):
        if "platform_name" in first_dataset.attrs:
            _platform_name = translate_platform_name.get(
                first_dataset.attrs["platform_name"],
                first_dataset.attrs["platform_name"])
        elif "platform_name" in kwargs:
            _platform_name = translate_platform_name.get(
                kwargs["platform_name"], kwargs["platform_name"])
        else:
            _platform_name = None
        return _platform_name

    def _calibrate_data(self, dataset, calibration, min_val, max_val):
        reverse_offset = 0.
        reverse_scale = 1.
        if calibration == "brightness_temperature":
            # If data is brightness temperature, the data must be inverted.
            reverse_offset = 255.
            reverse_scale = -1.
            dataset.data += KELVIN_TO_CELSIUS

        # Need to possible translate channels names from satpy to mitiff
        _data = reverse_offset + reverse_scale * ((dataset.data - float(min_val)) /
                                                  (float(max_val) - float(min_val))) * 255.
        return _data.clip(0, 255)

    def _save_as_palette(self, datasets, tmp_gen_filename, tiffinfo, **kwargs):
        # MITIFF palette has only one data channel
        if len(datasets.dims) == 2:
            LOG.debug("Palette ok with only 2 dimensions. ie only x and y")
            # 3 = Palette color. In this model, a color is described with a single component.
            # The value of the component is used as an index into the red, green and blue curves
            # in the ColorMap field to retrieve an RGB triplet that defines the color. When
            # PhotometricInterpretation=3 is used, ColorMap must be present and SamplesPerPixel must be 1.
            tiffinfo[270] = tiffinfo[270].decode("utf-8")

            img = Image.fromarray(datasets.data.astype(np.uint8), mode="P")
            if "palette_color_map" in kwargs:
                img.putpalette(ImagePalette.ImagePalette("RGB", kwargs["palette_color_map"]))
            else:
                LOG.error("In a mitiff palette image a color map must be provided: palette_color_map is missing.")
                return

            img.save(tmp_gen_filename, compression="raw", compress_level=9, tiffinfo=tiffinfo)

    def _save_as_enhanced(self, datasets, tmp_gen_filename, **kwargs):
        """Save datasets as an enhanced RGB image."""
        img = get_enhanced_image(datasets.squeeze(), enhance=self.enhancer)
        tiffinfo = {}
        if "bands" in img.data.sizes and "bands" not in datasets.sizes:
            LOG.debug("Datasets without 'bands' become image with 'bands' due to enhancement.")
            LOG.debug("Needs to regenerate mitiff image description")
        image_description = self._make_image_description(img.data, **kwargs)
        tiffinfo[IMAGEDESCRIPTION] = (image_description).encode("utf-8")

        mitiff_frames = []
        for band in img.data["bands"]:
            chn = img.data.sel(bands=band)
            data = chn.values.clip(0, 1) * 254. + 1
            data = np.nan_to_num(data, copy=False)
            data = data.clip(0, 255)
            mitiff_frames.append(Image.fromarray(data.astype(np.uint8), mode="L"))
        mitiff_frames[0].save(tmp_gen_filename, save_all=True, append_images=mitiff_frames[1:],
                              compression="raw", compress_level=9, tiffinfo=tiffinfo)

    def _generate_intermediate_filename(self, gen_filename):
        """Replace mitiff ext because pillow doesn't recognise the file type."""
        bs, ex = os.path.splitext(gen_filename)
        tmp_gen_filename = gen_filename
        if ex.endswith("mitiff"):
            bd = os.path.dirname(bs)
            bn = os.path.basename(bs)
            tmp_gen_filename = os.path.join(bd, "." + bn + ".tif")
        return tmp_gen_filename

    def _save_datasets_as_mitiff(self, datasets, image_description,
                                 gen_filename, **kwargs):
        """Put all together and save as a tiff file.

        Include the special tags making it a mitiff file.

        """
        tmp_gen_filename = self._generate_intermediate_filename(gen_filename)
        tiffinfo = {}
        tiffinfo[IMAGEDESCRIPTION] = (image_description).encode("latin-1")

        cns = self.translate_channel_name.get(kwargs["sensor"], {})
        if isinstance(datasets, list):
            LOG.debug("Saving datasets as list")
            mitiff_frames = []
            for _cn in self.channel_order[kwargs["sensor"]]:
                for dataset in datasets:
                    if dataset.attrs["name"] == _cn:
                        # Need to possible translate channels names from satpy to mitiff
                        cn = cns.get(dataset.attrs["name"], dataset.attrs["name"])
                        data = self._calibrate_data(dataset, dataset.attrs["calibration"],
                                                    self.mitiff_config[kwargs["sensor"]][cn]["min-val"],
                                                    self.mitiff_config[kwargs["sensor"]][cn]["max-val"])
                        mitiff_frames.append(Image.fromarray(data.astype(np.uint8), mode="L"))
                        break
            mitiff_frames[0].save(tmp_gen_filename, save_all=True, append_images=mitiff_frames[1:],
                                  compression="raw", compress_level=9, tiffinfo=tiffinfo)
        elif "dataset" in datasets.attrs["name"]:
            LOG.debug("Saving dataset as single dataset.")
            self._save_single_dataset(datasets, cns, tmp_gen_filename, tiffinfo, kwargs)
        elif self.palette:
            LOG.debug("Saving dataset as palette.")
            self._save_as_palette(datasets, tmp_gen_filename, tiffinfo, **kwargs)
        else:
            LOG.debug("Saving datasets as enhanced image")
            self._save_as_enhanced(datasets, tmp_gen_filename, **kwargs)
        os.rename(tmp_gen_filename, gen_filename)

    def _save_single_dataset(self, datasets, cns, tmp_gen_filename, tiffinfo, kwargs):
        LOG.debug("Saving %s as a dataset.", datasets.attrs["name"])
        if len(datasets.dims) == 2 and (all("bands" not in i for i in datasets.dims)):
            # Special case with only one channel ie. no bands

            # Need to possible translate channels names from satpy to mitiff
            # Note the last index is a tuple index.
            cn = cns.get(datasets.attrs["prerequisites"][0]["name"],
                         datasets.attrs["prerequisites"][0]["name"])
            data = self._calibrate_data(datasets, datasets.attrs["prerequisites"][0].get("calibration"),
                                        self.mitiff_config[kwargs["sensor"]][cn]["min-val"],
                                        self.mitiff_config[kwargs["sensor"]][cn]["max-val"])
            Image.fromarray(data.astype(np.uint8)).save(tmp_gen_filename, compression="raw",
                                                        compress_level=9, tiffinfo=tiffinfo)
        else:
            mitiff_frames = []
            for _cn_i, _cn in enumerate(self.channel_order[kwargs["sensor"]]):
                for band in datasets["bands"]:
                    if band == _cn:
                        chn = datasets.sel(bands=band)
                        # Need to possible translate channels names from satpy to mitiff
                        # Note the last index is a tuple index.
                        cn = cns.get(chn.attrs["prerequisites"][_cn_i]["name"],
                                     chn.attrs["prerequisites"][_cn_i]["name"])
                        data = self._calibrate_data(chn, chn.attrs["prerequisites"][_cn_i].get("calibration"),
                                                    self.mitiff_config[kwargs["sensor"]][cn]["min-val"],
                                                    self.mitiff_config[kwargs["sensor"]][cn]["max-val"])

                        mitiff_frames.append(Image.fromarray(data.astype(np.uint8), mode="L"))

                        break
            mitiff_frames[0].save(tmp_gen_filename, save_all=True, append_images=mitiff_frames[1:],
                                  compression="raw", compress_level=9, tiffinfo=tiffinfo)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for CF decoding."""

import datetime as dt

import pytest

import satpy.cf.decoding


class TestDecodeAttrs:
    """Test decoding of CF-encoded attributes."""

    @pytest.fixture
    def attrs(self):
        """Get CF-encoded attributes."""
        return {
            "my_integer": 0,
            "my_float": 0.0,
            "my_list": [1, 2, 3],
            "my_timestamp1": "2000-01-01",
            "my_timestamp2": "2000-01-01 12:15:33",
            "my_timestamp3": "2000-01-01 12:15:33.123456",
            "my_dict": '{"a": {"b": [1, 2, 3]}, "c": {"d": "2000-01-01 12:15:33.123456"}}'
        }

    @pytest.fixture
    def expected(self):
        """Get expected decoded results."""
        return {
            "my_integer": 0,
            "my_float": 0.0,
            "my_list": [1, 2, 3],
            "my_timestamp1": dt.datetime(2000, 1, 1),
            "my_timestamp2": dt.datetime(2000, 1, 1, 12, 15, 33),
            "my_timestamp3": dt.datetime(2000, 1, 1, 12, 15, 33, 123456),
            "my_dict": {"a": {"b": [1, 2, 3]},
                        "c": {"d": dt.datetime(2000, 1, 1, 12, 15, 33, 123456)}}
        }

    def test_decoding(self, attrs, expected):
        """Test decoding of CF-encoded attributes."""
        res = satpy.cf.decoding.decode_attrs(attrs)
        assert res == expected

    def test_decoding_doesnt_modify_original(self, attrs):
        """Test that decoding doesn't modify the original attributes."""
        satpy.cf.decoding.decode_attrs(attrs)
        assert isinstance(attrs["my_dict"], str)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for merging metadata from various sources."""

import datetime as dt
import warnings
from collections.abc import Collection
from functools import partial, reduce
from operator import eq, is_

import numpy as np

from satpy.utils import flatten_dict


def combine_metadata(*metadata_objects, average_times=None):
    """Combine the metadata of two or more Datasets.

    If the values corresponding to any keys are not equal or do not
    exist in all provided dictionaries then they are not included in
    the returned dictionary.

    All values of the keys containing the substring 'start_time' will be set
    to the earliest value and similarly for 'end_time' to latest time.  All
    other keys containing the word 'time' are averaged.  Before these adjustments,
    `None` values resulting from data that don't have times associated to them
    are removed. These rules are applied also to values in the 'time_parameters'
    dictionary.

    .. versionchanged:: 0.47

       Before Satpy 0.47, all times, including `start_time` and `end_time`, were averaged.

    In the interest of processing time, lazy arrays are compared by object
    identity rather than by their contents.

    Args:
        *metadata_objects: MetadataObject or dict objects to combine

    Kwargs:
        average_times (bool): Removed option to average all time attributes.

    Returns:
        dict: the combined metadata

    """
    if average_times is not None:
        warnings.warn(
            "'average_time' option has been removed and start/end times are handled with min/max instead.",
            UserWarning
        )

    info_dicts = _get_valid_dicts(metadata_objects)
    if len(info_dicts) == 1:
        return info_dicts[0].copy()

    shared_keys = _shared_keys(info_dicts)

    return _combine_shared_info(shared_keys, info_dicts)


def _get_valid_dicts(metadata_objects):
    """Get the valid dictionaries matching the metadata_objects."""
    info_dicts = []
    for metadata_object in metadata_objects:
        if isinstance(metadata_object, dict):
            metadata_dict = metadata_object
        elif hasattr(metadata_object, "attrs"):
            metadata_dict = metadata_object.attrs
        else:
            continue
        info_dicts.append(metadata_dict)
    return info_dicts


def _shared_keys(info_dicts):
    key_sets = (set(metadata_dict.keys()) for metadata_dict in info_dicts)
    return reduce(set.intersection, key_sets)


def _combine_shared_info(shared_keys, info_dicts):
    shared_info = {}
    for key in shared_keys:
        values = [info[key] for info in info_dicts]
        _combine_values(key, values, shared_info)
    return shared_info


def _combine_values(key, values, shared_info):
    if "time" in key:
        times = _combine_times(key, values)
        if times is not None:
            shared_info[key] = times
    elif _are_values_combinable(values):
        shared_info[key] = values[0]


def _combine_times(key, values):
    if key == "time_parameters":
        return _combine_time_parameters(values)
    filtered_values = _filter_time_values(values)
    if not filtered_values:
        return None
    if "end_time" in key:
        return max(filtered_values)
    elif "start_time" in key:
        return min(filtered_values)
    return average_datetimes(filtered_values)


def _combine_time_parameters(values):
    # Assume the first item has all the keys
    keys = values[0].keys()
    res = {}
    for key in keys:
        sub_values = [itm[key] for itm in values]
        res[key] = _combine_times(key, sub_values)
    return res


def _filter_time_values(values):
     """Remove values that are not datetime objects."""
     return [v for v in values if isinstance(v, dt.datetime)]


def average_datetimes(datetime_list):
    """Average a series of datetime objects.

    .. note::

        This function assumes all datetime objects are naive and in the same
        time zone (UTC).

    Args:
        datetime_list (Iterable): Datetime objects to average

    Returns: Average datetime as a datetime object

    """
    total = [dt.datetime.timestamp(d) for d in datetime_list]
    return dt.datetime.fromtimestamp(sum(total) / len(total))


def _are_values_combinable(values):
    """Check if the *values* can be combined."""
    if _contain_dicts(values):
        return _all_dicts_equal(values)
    return _all_non_dicts_equal(values)


def _all_non_dicts_equal(values):
    if _contain_arrays(values):
        return _all_arrays_equal(values)
    if _contain_collections_of_arrays(values):
        # in the real world, the `ancillary_variables` attribute may be
        # List[xarray.DataArray], this means our values are now
        # List[List[xarray.DataArray]].
        # note that this list_of_arrays check is also true for any
        # higher-dimensional ndarray, but we only use this check after we have
        # checked any_arrays so this false positive should have no impact
        return _all_list_of_arrays_equal(values)
    return _all_values_equal(values)


def _contain_arrays(values):
    return any([_is_array(value) for value in values])


def _is_array(val):
    """Check if val is an array."""
    return hasattr(val, "__array__") and not np.isscalar(val)


def _contain_dicts(values):
    return any(isinstance(value, dict) for value in values)


nan_allclose = partial(np.allclose, equal_nan=True)


def _all_arrays_equal(arrays):
    """Check if the arrays are equal.

    If the arrays are lazy, just check if they have the same identity.
    """
    if getattr(arrays[0], "chunks", None) is not None:
        return _all_identical(arrays)

    return _all_values_equal(arrays)


def _all_values_equal(values):
    try:
        return _all_close(values)
    except (ValueError, TypeError):
        # In case of object type arrays (e.g. datetime) _all_close fails,
        # but _all_equal succeeds.
        return _all_equal(values)


def _all_dicts_equal(dicts):
    try:
        return _pairwise_all(_dict_equal, dicts)
    except AttributeError:
        # There is something else than a dictionary in the list
        return False


def _dict_equal(d1, d2):
    """Check that two dictionaries are equal.

    Nested dictionaries are flattened to facilitate comparison.
    """
    d1_flat = flatten_dict(d1)
    d2_flat = flatten_dict(d2)
    if not _dict_keys_equal(d1_flat, d2_flat):
        return False
    for key in d1_flat.keys():
        value_pair = [d1_flat[key], d2_flat[key]]
        if not _all_non_dicts_equal(value_pair):
            return False
    return True


def _dict_keys_equal(d1, d2):
    return d1.keys() == d2.keys()


def _pairwise_all(func, values):
    for value in values[1:]:
        if not _is_equal(values[0], value, func):
            return False
    return True


def _is_equal(a, b, comp_func):
    res = comp_func(a, b)
    if _is_array(res):
        return res.all()
    return res


def _all_identical(values):
    """Check that the identities of all values are the same."""
    return _pairwise_all(is_, values)


def _all_close(values):
    return _pairwise_all(nan_allclose, values)


def _all_equal(values):
    return _pairwise_all(eq, values)


def _contain_collections_of_arrays(values):
    return any(
        [_is_non_empty_collection(value) and
         _is_all_arrays(value)
         for value in values])


def _is_non_empty_collection(value):
    return isinstance(value, Collection) and len(value) > 0


def _is_all_arrays(value):
    return all([_is_array(sub_value) for sub_value in value])


def _all_list_of_arrays_equal(array_lists):
    """Check that the lists of arrays are equal."""
    for array_list in zip(*array_lists):
        if not _all_arrays_equal(array_list):
            return False
    return True



#!/usr/bin/env python
# Copyright (c) 2018-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for compositors using auxiliary data."""

import os
import sys
import unittest
from unittest import mock

import pytest

import satpy

if sys.platform.startswith("win"):
    IMAGE_FILENAME = r"C:\foo.tif"
    PATH_TO_IMAGE = r"C:\path\to\image"
else:
    IMAGE_FILENAME = "/foo.tif"
    PATH_TO_IMAGE = "/path/to/image"

class TestStaticImageCompositor(unittest.TestCase):
    """Test case for the static compositor."""

    @mock.patch("satpy.area.get_area_def")
    def test_init(self, get_area_def):
        """Test the initializiation of static compositor."""
        from satpy.composites.aux_data import StaticImageCompositor

        # No filename given raises ValueError
        with pytest.raises(ValueError, match="StaticImageCompositor needs a .*"):
            StaticImageCompositor("name")

        # No area defined
        comp = StaticImageCompositor("name", filename=IMAGE_FILENAME)
        assert comp._cache_filename == IMAGE_FILENAME
        assert comp.area is None

        # Area defined
        get_area_def.return_value = "bar"
        comp = StaticImageCompositor("name", filename=IMAGE_FILENAME, area="euro4")
        assert comp._cache_filename == IMAGE_FILENAME
        assert comp.area == "bar"
        get_area_def.assert_called_once_with("euro4")

    @mock.patch("satpy.aux_download.retrieve")
    @mock.patch("satpy.aux_download.register_file")
    @mock.patch("satpy.Scene")
    def test_call(self, Scene, register, retrieve):  # noqa
        """Test the static compositing."""
        from satpy.composites.aux_data import StaticImageCompositor

        satpy.config.set(data_dir=PATH_TO_IMAGE)
        remote_tif = "http://example.com/foo.tif"

        class MockScene(dict):
            def load(self, arg):
                pass

        img = mock.MagicMock()
        img.attrs = {}
        scn = MockScene()
        scn["image"] = img
        Scene.return_value = scn
        # absolute path to local file
        comp = StaticImageCompositor("name", filename=IMAGE_FILENAME, area="euro4")
        res = comp()
        Scene.assert_called_once_with(reader="generic_image",
                                      filenames=[IMAGE_FILENAME])
        register.assert_not_called()
        retrieve.assert_not_called()
        assert res.attrs["sensor"] is None
        assert "modifiers" not in res.attrs
        assert "calibration" not in res.attrs

        # remote file with local cached version
        Scene.reset_mock()
        register.return_value = os.path.join("data_dir", "foo.tif")
        retrieve.return_value = os.path.join("data_dir", "foo.tif")
        comp = StaticImageCompositor("name", url=remote_tif, area="euro4")
        res = comp()
        Scene.assert_called_once_with(reader="generic_image",
                                      filenames=[os.path.join("data_dir", "foo.tif")])
        assert res.attrs["sensor"] is None
        assert "modifiers" not in res.attrs
        assert "calibration" not in res.attrs

        # Non-georeferenced image, no area given
        img.attrs.pop("area")
        comp = StaticImageCompositor("name", filename=IMAGE_FILENAME)
        with pytest.raises(AttributeError):
            comp()

        # Non-georeferenced image, area given
        comp = StaticImageCompositor("name", filename=IMAGE_FILENAME, area="euro4")
        res = comp()
        assert res.attrs["area"].area_id == "euro4"

        # Filename contains environment variable
        os.environ["TEST_IMAGE_PATH"] = PATH_TO_IMAGE
        comp = StaticImageCompositor("name", filename=os.path.join("${TEST_IMAGE_PATH}", "foo.tif"), area="euro4")
        assert comp._cache_filename == os.path.join(PATH_TO_IMAGE, "foo.tif")

        # URL and filename without absolute path
        comp = StaticImageCompositor("name", url=remote_tif, filename="bar.tif")
        assert comp._url == remote_tif
        assert comp._cache_filename == "bar.tif"

        # No URL, filename without absolute path, use default data_dir from config
        with mock.patch("os.path.exists") as exists:
            exists.return_value = True
            comp = StaticImageCompositor("name", filename="foo.tif")
            assert comp._url is None
            assert comp._cache_filename == os.path.join(PATH_TO_IMAGE, "foo.tif")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the Multiscene object."""

import unittest
from unittest import mock

import pytest
import xarray as xr

from satpy import DataQuery
from satpy.tests.multiscene_tests.test_utils import _create_test_area, _create_test_dataset, _create_test_scenes
from satpy.tests.utils import make_dataid


class TestMultiScene(unittest.TestCase):
    """Test basic functionality of MultiScene."""

    def test_init_empty(self):
        """Test creating a multiscene with no children."""
        from satpy import MultiScene
        MultiScene()

    def test_init_children(self):
        """Test creating a multiscene with children."""
        from satpy import MultiScene
        scenes = _create_test_scenes()
        MultiScene(scenes)

    def test_properties(self):
        """Test basic properties/attributes of the MultiScene."""
        from satpy import MultiScene

        area = _create_test_area()
        scenes = _create_test_scenes(area=area)
        ds1_id = make_dataid(name="ds1")
        ds2_id = make_dataid(name="ds2")
        ds3_id = make_dataid(name="ds3")
        ds4_id = make_dataid(name="ds4")

        # Add a dataset to only one of the Scenes
        scenes[1]["ds3"] = _create_test_dataset("ds3")
        mscn = MultiScene(scenes)

        assert mscn.loaded_dataset_ids == {ds1_id, ds2_id, ds3_id}
        assert mscn.shared_dataset_ids == {ds1_id, ds2_id}
        assert mscn.all_same_area

        bigger_area = _create_test_area(shape=(20, 40))
        scenes[0]["ds4"] = _create_test_dataset("ds4", shape=(20, 40),
                                                area=bigger_area)

        assert mscn.loaded_dataset_ids == {ds1_id, ds2_id, ds3_id, ds4_id}
        assert mscn.shared_dataset_ids == {ds1_id, ds2_id}
        assert not mscn.all_same_area

    def test_from_files(self):
        """Test creating a multiscene from multiple files."""
        from satpy import MultiScene
        input_files_abi = [
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171502203_e20171171504576_c20171171505018.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171507203_e20171171509576_c20171171510018.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171512203_e20171171514576_c20171171515017.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171517203_e20171171519577_c20171171520019.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171522203_e20171171524576_c20171171525020.nc",
            "OR_ABI-L1b-RadC-M3C01_G16_s20171171527203_e20171171529576_c20171171530017.nc",
        ]
        input_files_glm = [
            "OR_GLM-L2-GLMC-M3_G16_s20171171500000_e20171171501000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171501000_e20171171502000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171502000_e20171171503000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171503000_e20171171504000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171504000_e20171171505000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171505000_e20171171506000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171506000_e20171171507000_c20380190314080.nc",
            "OR_GLM-L2-GLMC-M3_G16_s20171171507000_e20171171508000_c20380190314080.nc",
        ]
        with mock.patch("satpy.multiscene._multiscene.Scene") as scn_mock:
            mscn = MultiScene.from_files(
                input_files_abi,
                reader="abi_l1b",
                scene_kwargs={"reader_kwargs": {}})
            assert len(mscn.scenes) == 6
            calls = [mock.call(
                filenames={"abi_l1b": [in_file_abi]},
                reader_kwargs={})
                for in_file_abi in input_files_abi]
            scn_mock.assert_has_calls(calls)

            scn_mock.reset_mock()
            with pytest.warns(DeprecationWarning, match=".*ensure_all_readers.*"):
                mscn = MultiScene.from_files(
                    input_files_abi + input_files_glm,
                    reader=("abi_l1b", "glm_l2"),
                    group_keys=["start_time"],
                    ensure_all_readers=True,
                    time_threshold=30)
            assert len(mscn.scenes) == 2
            calls = [mock.call(
                filenames={"abi_l1b": [in_file_abi], "glm_l2": [in_file_glm]})
                for (in_file_abi, in_file_glm) in
                zip(input_files_abi[0:2],
                    [input_files_glm[2]] + [input_files_glm[7]])]
            scn_mock.assert_has_calls(calls)
            scn_mock.reset_mock()
            mscn = MultiScene.from_files(
                input_files_abi + input_files_glm,
                reader=("abi_l1b", "glm_l2"),
                group_keys=["start_time"],
                ensure_all_readers=False,
                time_threshold=30)
            assert len(mscn.scenes) == 12


class TestMultiSceneGrouping:
    """Test dataset grouping in MultiScene."""

    @pytest.fixture
    def scene1(self):
        """Create first test scene."""
        from satpy import Scene
        scene = Scene()
        dsid1 = make_dataid(
            name="ds1",
            resolution=123,
            wavelength=(1, 2, 3),
            polarization="H"
        )
        scene[dsid1] = _create_test_dataset(name="ds1")
        dsid2 = make_dataid(
            name="ds2",
            resolution=456,
            wavelength=(4, 5, 6),
            polarization="V"
        )
        scene[dsid2] = _create_test_dataset(name="ds2")
        return scene

    @pytest.fixture
    def scene2(self):
        """Create second test scene."""
        from satpy import Scene
        scene = Scene()
        dsid1 = make_dataid(
            name="ds3",
            resolution=123.1,
            wavelength=(1.1, 2.1, 3.1),
            polarization="H"
        )
        scene[dsid1] = _create_test_dataset(name="ds3")
        dsid2 = make_dataid(
            name="ds4",
            resolution=456.1,
            wavelength=(4.1, 5.1, 6.1),
            polarization="V"
        )
        scene[dsid2] = _create_test_dataset(name="ds4")
        return scene

    @pytest.fixture
    def multi_scene(self, scene1, scene2):
        """Create small multi scene for testing."""
        from satpy import MultiScene
        return MultiScene([scene1, scene2])

    @pytest.fixture
    def groups(self):
        """Get group definitions for the MultiScene."""
        return {
            DataQuery(name="odd"): ["ds1", "ds3"],
            DataQuery(name="even"): ["ds2", "ds4"]
        }

    def test_multi_scene_grouping(self, multi_scene, groups, scene1):
        """Test grouping a MultiScene."""
        multi_scene.group(groups)
        shared_ids_exp = {make_dataid(name="odd"), make_dataid(name="even")}
        assert multi_scene.shared_dataset_ids == shared_ids_exp
        assert DataQuery(name="odd") not in scene1
        xr.testing.assert_allclose(multi_scene.scenes[0]["ds1"], scene1["ds1"])

    def test_fails_to_add_multiple_datasets_from_the_same_scene_to_a_group(self, multi_scene):
        """Test that multiple datasets from the same scene in one group fails."""
        groups = {DataQuery(name="mygroup"): ["ds1", "ds2"]}
        multi_scene.group(groups)
        with pytest.raises(ValueError, match="Cannot add multiple datasets from a scene to the same group"):
            next(multi_scene.scenes)
